/*
    Persistence v1.04 is a easy-use O-R and O-F frame which suit local software.   
    Copyright (C) 2009 www.nonesole.com Jack Lee

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package com.nonesole.persistence.connectionpool;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.UnknownHostException;
import java.security.AccessControlException;

import com.nonesole.persistence.exception.OperationsException;
import com.nonesole.persistence.tools.DateToolkit;
import com.nonesole.persistence.tools.MD5;

/**
 * Connection Pool Server.
 * <pre>
 *   Service: 
 *   1.start a non-stop thread with connection pool
 *   2.close the non-stop thread
 *   3.initialize a connection pool without thread
 * </pre>
 * @author JACK LEE
 * @version 1.01 - modify in 2009-08-06 changed the method of creating Close_key
 * @version 1.0 - build in 2009-07-08
 */
public class ConnectionPoolServer {
	
	//Key of closing thread
	private final static String CLOSE_KEY;
	private static boolean isServerStart = false;
	
	

	static {
		CLOSE_KEY = MD5.encrypt(String.valueOf(Math.random()));
	}
	
	/**
	 * Connection Pool Manager
	 * */
	private static SingleRecipientManager srm;
	
	/**
	 * Non-stop thread with connection pool
	 * */
	private static Thread currentThread;
	
	/**
	 * Initialize connection pool without thread
	 * @throws OperationsException 
	 * */
	public final static void initPool() throws OperationsException{
		srm = SingleRecipientManager.getInstance();
		srm.start();
	}
		
	/**
	 * start a non-stop thread with connection pool
	 * @throws OperationsException 
	 * */
	public final static void start() throws OperationsException{
		if ( null != currentThread )
			System.err.println("Cannot start server," +
						"because server has been started.");
		if ( null != srm )
			System.err.println("Failure.The pool has been exist.");
		initPool();
		System.out.println("Server start at "+ DateToolkit.now() +".");
		currentThread = new SocketedThread();
		currentThread.start();
		isServerStart = true;
	}

	/**
	 * Clean current thread
	 * */
	private static void clean(){
		srm = null;
		currentThread = null;
	}
	
	/**
	 * close the non-stop thread
	 * @throws OperationsException 
	 * */
	public final static void stop(){		
		try {
			Socket socket = new Socket(LOCAL_IP,PORT);
			OutputStream os = socket.getOutputStream();
			os.write(CLOSE_KEY.getBytes());
			os.flush();
		} catch (UnknownHostException e) {
			 System.err.println("Client unkonwHost exception: "
                     + e.getMessage());
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println("Client io exception: "
                    + e.getMessage());
			e.printStackTrace();
			System.exit(1);
		} finally {
			isServerStart = false;
		}
	}
	
	/**
	 * Local IP
	 * */
	private static final String LOCAL_IP = "127.0.0.1";
	
	/**
	 * Socket Port
	 * */
	private static final int PORT = 600;
	
	/**
	 * Socket thread ,it will keep connection pool start all the time.
	 * Loop waiting for key which would end the thread.
	 * */
	private static class SocketedThread extends Thread {
		public void run(){
			ServerSocket serverSocket = null;
			try {
	            serverSocket =
	                new ServerSocket(PORT, 1,
	                                 InetAddress.getByName(LOCAL_IP));
	        } catch (IOException e) {
	            System.err.println("await: create[" + PORT + "]: " + e);
	            e.printStackTrace();
	            System.exit(1);
	        }
	        
	        // Loop waiting for a connection and a valid command
	        Socket socket = null;
	        byte[] key = new byte[CLOSE_KEY.length()];
	        while (true) {
	            // Wait for the next connection
	            InputStream stream = null;
	            try {		            	
	                socket = serverSocket.accept();
	                socket.setSoTimeout(10 * 1000);  // Ten seconds
	                stream = socket.getInputStream();
	                stream.read(key);
	                if ( CLOSE_KEY.equals(new String(key)) )  {
	                	// Close the socket now that we are done with it
	                	try {
	    	                socket.close();
	    	                serverSocket.close();
	    	                srm.close();
	    	                clean();
	    	                System.out.println(
	    	                		"Server is closed at "+ DateToolkit.now() +".");
	    	            } catch (IOException e) {
	    	                ;
	    	            } catch (OperationsException e) {
	    					;
	    				} finally {
	    	            	isServerStart = false;
	    	            }
	    	            return;
	                }
	            } catch (AccessControlException ace) {
	                System.err.println("Server accept security exception: "
	                                   + ace.getMessage());
	                continue;
	            } catch (IOException e) {
	                System.err.println("Server accept: " + e);
	                e.printStackTrace();
	                System.exit(1);
	            } finally {
	            	isServerStart = false;
	            }
	        }
		}
	}
	
	/**
	 * @return the isServerStart
	 */
	public static boolean isServerStart() {
		return isServerStart;
	}
}
