/*
    Persistence v1.04 is a easy-use O-R and O-F frame which suit local software.   
    Copyright (C) 2009 www.nonesole.com Jack Lee

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package com.nonesole.persistence.connectionpool;

import java.io.PrintWriter;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;

import javax.sql.DataSource;

import com.nonesole.persistence.connectionpool.parser.DriverProperties;
import com.nonesole.persistence.exception.OperationsException;

/**
 * Local data source
 * <pre>
 * Use
 * 1.Create new datasource from jdbc config
 * &#9;	DriverDataSource dds = new DriverDataSource(new DriverProperties());
 * 2.Create new datasource from context of server just like Tomcat
 * &#9;	Context ctx = new InitialContext();
 * &#9;	DataSource ds = ctx.lookup("jdbc/db");
 * 3.Create new datasource from Persistence frame
 * &#9;	ConnectionPoolServer.initPool();
 * &#9;	DataSource ds = ConnectionPoolContext.getDataSource("mysql");
 * </pre>
 * @author JACK LEE
 * @version 1.0 - build in 2009-07-04
 */
public class DriverDataSource implements DataSource {

	//jdbc config object
	private DriverProperties driver;
	//Connection object
	private Connection conn;
	//Exception head
	private final static String REPORTER = "DriverDataSource ";
	private final static String KEY_ENTER = "\r\n";
		
	/**
	 * Construction 
	 * @param driver jdbc config object 
	 * @throws OperationsException 
	 * */
	public DriverDataSource(DriverProperties driver) throws OperationsException{
		this.driver = driver;
		if( !this.driver.isValid() )
			throw new OperationsException("DataSource Config is failure.");
	}
		
	/**
	 * Connection Pool manager
	 * */
	private SingleRecipientManager srm = null;
	
	/**
	 * Set Connection Pool manager 
	 * @param srm Connecton Pool manager
	 * */
	protected void setRecipientManager(SingleRecipientManager srm){
		this.srm = srm;
	}
	
	/**
	 * Initialize Connection
	 * @throws OperationsException 
	 * */
	private synchronized final void initConnection() throws OperationsException {
		try {
			Class.forName(this.driver.getDriverClass()).newInstance();
			conn = DriverManager.getConnection(
					driver.getUrl(),driver.getUser(),driver.getPwd());
		} catch (InstantiationException e) {
			throwExcpetion(REPORTER + e.getMessage() + KEY_ENTER,
					e.fillInStackTrace());
		} catch (IllegalAccessException e) {
			throwExcpetion(REPORTER + e.getMessage() + KEY_ENTER,
					e.fillInStackTrace());
		} catch (ClassNotFoundException e) {
			throwExcpetion(REPORTER + e.getMessage() + KEY_ENTER,
					e.fillInStackTrace());
		} catch (SQLException e) {
			throwExcpetion(REPORTER + e.getMessage() + KEY_ENTER,
					e.fillInStackTrace());
		}
	}
	
	/* (non-Javadoc)
	 * @see javax.sql.DataSource#getConnection()
	 */
	public synchronized Connection getConnection() throws SQLException {
		return getConnection(null,null);
	}

	/* (non-Javadoc)
	 * @see javax.sql.DataSource#getConnection(java.lang.String, java.lang.String)
	 */
	public synchronized Connection getConnection(String username, String password)
			throws SQLException {
		if ( null != username )
			driver.setUser(username);
		if ( null != password )
			driver.setPwd(password);
		
		//Check the configuration whether is valid
		if ( !driver.isValid() )
			throw new SQLException("Connection config is failure.");
		
		if ( driver.isAssociateConnectionPool() ) {
			if ( null == srm  )
				throw new SQLException("Connection Pool is null.");
			try {
				return srm.getConnection(driver.getDataSourceName());
			} catch (OperationsException e) {
				throw new SQLException(
						"Exception : " + e.getMessage() + 
						" Cause by " + e.getCause());
			} catch (InterruptedException e) {
				throw new SQLException(
						"Getting Connection is failure." + e.getMessage() + 
						" Cause by " + e.getCause());
			}
		} else {//directly connect database with jdbc
			try {
				initConnection();
			} catch (OperationsException e) {
				throw new SQLException(
						"Exception : " + e.getMessage() + 
						" Cause by " + e.getCause());
			}
			return conn;
		}
	}
	
	/**
	 * ׳־ûָ쳣
	 * @param message String 
	 * @param cause Throwable
	 * */
	private void throwExcpetion(String message,Throwable cause) throws OperationsException{
		throw new OperationsException(message,cause);
	}
	
	private PrintWriter writer;
	
	/* (non-Javadoc)
	 * @see javax.sql.DataSource#getLogWriter()
	 */
	public PrintWriter getLogWriter() throws SQLException {
		return writer;
	}
	
	/* (non-Javadoc)
	 * @see javax.sql.DataSource#setLogWriter(java.io.PrintWriter)
	 */
	public void setLogWriter(PrintWriter out) throws SQLException {
		writer = out;
	}

	private int maxLoginTime = 0; 

	/* (non-Javadoc)
	 * @see javax.sql.DataSource#getLoginTimeout()
	 */
	public int getLoginTimeout() throws SQLException {
		return maxLoginTime;
	}
	
	/* (non-Javadoc)
	 * @see javax.sql.DataSource#setLoginTimeout(int)
	 */
	public void setLoginTimeout(int seconds) throws SQLException {
		maxLoginTime = seconds;
	}

	/**
	 * Get driver configuration
	 * @see com.nonesole.persistence.connectionpool.IDriverDataSource#getDriverProperties()
	 * */
	public DriverProperties getDriverProperties() {
		return driver;
	}

}
