/*
    Persistence v1.04 is a easy-use O-R and O-F frame which suit local software.   
    Copyright (C) 2009 www.nonesole.com Jack Lee

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package com.nonesole.persistence.connectionpool;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import javax.sql.ConnectionEvent;
import javax.sql.ConnectionEventListener;
import javax.sql.PooledConnection;


/**
 * Pooled connection class.
 * @author JACK LEE
 * @version 1.0 - build in 2009-07-05
 */
public class DriverPooledConnection implements PooledConnection {

	/**
	 * Close normally
	 * */
	public static final int CONNECTION_CLOSED = 0;
	
	/**
	 * Close by error
	 * */
	public static final int CONNECTION_ERROR_OCCURRED = 1;
	
	/**
	 * Manage connection event listeners 
	 * */
	private Map<Object,Object> connectionEventListeners;
	
	/**
	 * When pooled connection is closed,
	 * the one which is closed is logic connection.
	 * the real physical connection will return 
	 * to connection pool.
	 * */
	private Connection logicalHandle;

	/**
	 * physical connection
	 * */
	private Connection physicalConn;

	/**
	 * Construction
	 * @param physical connection
	 * */
	protected DriverPooledConnection(Connection physicalConn) {
		this.physicalConn = physicalConn;
	}

	/* (non-Javadoc)
	 * @see javax.sql.PooledConnection#close()
	 */
	public synchronized void close() throws SQLException {
		if (this.physicalConn != null) {
			this.physicalConn.close();
			this.physicalConn = null;
		}
		
		if (this.connectionEventListeners != null) {
			this.connectionEventListeners.clear();
			this.connectionEventListeners = null;
		}
	}

	/* (non-Javadoc)
	 * @see javax.sql.PooledConnection#getConnection()
	 */
	public synchronized Connection getConnection() throws SQLException {
		
		if (this.physicalConn == null) {//check physical connection
			
			SQLException sqlException = new SQLException(
					"Physical Connection doesn't exist");
			stateChangeNotice(CONNECTION_CLOSED, sqlException);

			throw sqlException;
		}

		try {

			if (this.logicalHandle != null) {//close logical connection
				((ConnectionWrapper) this.logicalHandle).close();
			}
			
			//release physical connection
			this.physicalConn.clearWarnings();
			//this.physicalConn.rollback();
			this.physicalConn.setAutoCommit(true);			
			this.logicalHandle = ConnectionWrapper.getInstance(this, this.physicalConn);
		} catch (SQLException sqlException) {
			stateChangeNotice(CONNECTION_ERROR_OCCURRED, sqlException);
			throw sqlException;
		}

		return this.logicalHandle;
	}

	/* (non-Javadoc)
	 * @see javax.sql.PooledConnection#addConnectionEventListener(javax.sql.ConnectionEventListener)
	 */
	public synchronized void addConnectionEventListener(ConnectionEventListener listener) {
		if (this.connectionEventListeners != null) {
			this.connectionEventListeners.put(listener,
					listener);
		}
	}
	
	/* (non-Javadoc)
	 * @see javax.sql.PooledConnection#removeConnectionEventListener(javax.sql.ConnectionEventListener)
	 */
	public synchronized void removeConnectionEventListener(ConnectionEventListener listener) {
		if (this.connectionEventListeners != null) {
			this.connectionEventListeners.remove(listener);
		}
	}
	
	/**
	 * To remind the state of change.
	 * @param eventType
	 * @param sqlException
	 * */
	protected synchronized void stateChangeNotice(int eventType,SQLException sqlException) {
		if (this.connectionEventListeners == null)
			return;
		Iterator<Entry<Object, Object>> iterator = this.connectionEventListeners.entrySet().iterator();
		ConnectionEvent connectionevent = new ConnectionEvent(this,sqlException);
		while (iterator.hasNext()) {
			ConnectionEventListener connectioneventlistener = 
				(ConnectionEventListener) iterator.next().getValue();
			if (eventType == CONNECTION_CLOSED) {
				connectioneventlistener.connectionClosed(connectionevent);
			} else if (eventType == CONNECTION_ERROR_OCCURRED) {
				connectioneventlistener
						.connectionErrorOccurred(connectionevent);
			}
		}
	}
	
}
