/*
    Persistence v1.04 is a easy-use O-R and O-F frame which suit local software.   
    Copyright (C) 2009 www.nonesole.com Jack Lee

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package com.nonesole.persistence.connectionpool;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import javax.naming.NamingException;
import javax.sql.ConnectionEvent;
import javax.sql.ConnectionEventListener;
import javax.sql.PooledConnection;

import com.nonesole.persistence.connectionpool.parser.ConnectionPoolConfigParser;
import com.nonesole.persistence.connectionpool.parser.ConnectionPoolProperties;
import com.nonesole.persistence.connectionpool.parser.DataSourceConfigParser;
import com.nonesole.persistence.connectionpool.parser.DriverProperties;
import com.nonesole.persistence.exception.OperationsException;

/**
 * Manager of pool.<br>
 * Support one pool different DataSources.
 * <pre>
 * Use
 * 	   SingleRecipientManager.getInstance().start()
 * </pre>
 * @author JACK LEE
 * @version 1.0 - build in 2009-07-01
 */
public class SingleRecipientManager {

	//list of JDBC configuration object
	private List<DriverProperties> DRIVERS;
	//manager of pool	
	private static final SingleRecipientManager MANAGER;
	//pooled connection map,key is DataSource name
	private Map<String,Vector<PooledConnection>> POOLED_CONNECTION_MAP; 
	//physical connection map,key is DataSource name
	private Map<String,ConnectionPoolProperties> CONNECTION_POOL_PROPERTY_MAP;
	//various JDBC configurations of DataSource, key is DataSource name
	private Map<String,DriverProperties> DRIVER_PROPERTY_MAP;
	//pooled connections which have been used , key is DataSource name
	private Map<String,Vector<PooledConnection>> USED_POOLED_CONNECTION_MAP; 
	//waiting time of thread
	private static int WAIT_MIL_SECONDES = 0;
	//Map of the manager,it is for Multi-threaded manage,  key is DataSource name
	private Map<String,Vector<SingleRecipientManager>> CONTAINER_MAP;
	
	private Thread currentThread;
	
	private SingleRecipientManager(){}
	
	static {
		MANAGER = new SingleRecipientManager();
	}
	
	/**
	 * Return SingleRecipientManager object
	 * @return SingleRecipientManager Object
	 * */
	protected static final SingleRecipientManager getInstance(){
		return MANAGER;
	}
	
	/**
	 * close manager
	 * */
	protected final void close() throws OperationsException {
		CONTAINER_MAP = null;
		USED_POOLED_CONNECTION_MAP = null;
		POOLED_CONNECTION_MAP = null;
		CONNECTION_POOL_PROPERTY_MAP = null;
		DRIVER_PROPERTY_MAP = null;
		DRIVERS = null;
		currentThread = null;
	}
	
	/**
	 * start manager.
	 * <pre>
	 * 1.initialize every parameters of connection pool
	 * 2.parse xml files and import configuration of database and connection pool
	 * 3.initialize connection pool,bind DataSource and start a daemon thread
	 * </pre>
	 * @throws OperationsException
	 * */
	protected final void start() throws OperationsException{
		try {
			CONTAINER_MAP = new HashMap<String,Vector<SingleRecipientManager>>();
			USED_POOLED_CONNECTION_MAP = new HashMap<String,Vector<PooledConnection>>();
			POOLED_CONNECTION_MAP = new HashMap<String,Vector<PooledConnection>>();
			CONNECTION_POOL_PROPERTY_MAP = new HashMap<String,ConnectionPoolProperties>();
			DRIVER_PROPERTY_MAP = new HashMap<String,DriverProperties>();
			DRIVERS = getEffectiveDrivers(DataSourceConfigParser.getDataSources());
			initConectionPools();
		} catch (SQLException e) {
			throwExcpetion("SingleRecipientManager start failure.",e.fillInStackTrace());			
		} catch (OperationsException e) {
			throwExcpetion("SingleRecipientManager start failure.",e.fillInStackTrace());
		} catch (NamingException e) {
			throwExcpetion("SingleRecipientManager start failure.",e.fillInStackTrace());
		}
	}
	
	/**
	 * Pick up effective JDBC configuration drivers for DataSource.
	 * @param l list of JDBC configuration drivers from configuration files
	 * @return effective JDBC configuration drivers
	 * @throws OperationsException 
	 * */
	private final List<DriverProperties> getEffectiveDrivers(List<DriverProperties> l) throws OperationsException{
		List<DriverProperties> d = new ArrayList<DriverProperties>();
		Set<String> poolConfig = new HashSet<String>();
		for ( DriverProperties dp : l ) {
			if ( dp.isAssociateConnectionPool() ) {
				poolConfig.add(dp.getConnectionPoolName());
				if ( poolConfig.size() > 1 )
					throw new OperationsException(
							"This container allows only one pool manager.The config is failure. ");
				d.add(dp);
			}
		}
		l = null;
		return d;
	}
	
	/**
	 * Initialize connection pools
	 * @throws SQLException 
	 * @throws OperationsException 
	 * @throws NamingException 
	 * */
	private void initConectionPools() 
	throws SQLException, OperationsException, NamingException{
		String dataSourceName = "";
		String poolName = "";
		Vector<PooledConnection> pcv;
		ConnectionPoolProperties cpp;
		DriverDataSource daraSourceObj;
		
		for ( DriverProperties dp : DRIVERS ) {
			if ( !dp.isValid() )
				throw new OperationsException("Driver " + dp.getDataSourceName() + " Config is failure.");
			dataSourceName = dp.getDataSourceName();
			poolName = dp.getConnectionPoolName();
			pcv = new Vector<PooledConnection>();
			cpp = ConnectionPoolConfigParser.getConnectionPool(poolName);
			
			daraSourceObj = new DriverDataSource(dp);
			daraSourceObj.setRecipientManager(MANAGER);
			ConnectionPoolContext
				.getInstance()
				.getInitialContext()
				.bind(dataSourceName,daraSourceObj);//bind DataSource at JNDI
			
			POOLED_CONNECTION_MAP.put(dataSourceName, pcv);
			DRIVER_PROPERTY_MAP.put(dataSourceName, dp);
			CONNECTION_POOL_PROPERTY_MAP.put(dataSourceName, cpp);
			USED_POOLED_CONNECTION_MAP.put(
					dataSourceName, new Vector<PooledConnection>());

			CONTAINER_MAP.put(dataSourceName, new Vector<SingleRecipientManager>());
			
			currentThread = new PoolManageThread(dataSourceName);//start daemon thread 
			currentThread.start();
			
			createConnectionPool(
					DriverConnectionPoolDataSource.getInstance(dp),
					cpp,
					pcv,
					dataSourceName);
			
		}
	}
		
	/**
	 * Create connection pool
	 * @param dcpd DriverConnectionPoolDataSource object
	 * @param cpp ConnectionPoolProperties object
	 * @param cpp list of pooled connection
	 * @param DataSource Name
	 * @throws OperationsException 
	 * */
	private void createConnectionPool(
			DriverConnectionPoolDataSource dcpd,
			ConnectionPoolProperties cpp,
			Vector<PooledConnection> pcv,
			String dataSourceName) throws SQLException, OperationsException{
		int initSize = cpp.getInitialPoolSize();//pool size
		PooledConnection pc = dcpd.getPooledConnection();
		pc.addConnectionEventListener(
				new PooledConnectionListener(dataSourceName));
		
		if ( !cpp.isValid() )
			throw new OperationsException(
					"Pool " + cpp.getPoolConfigName() +" Config is failure.");
		
		for ( int i = 0 ; i < initSize ; i++ ) {
			pcv.add(pc);
		}
	}
	
	/**
	 * Get connection by DataSource name
	 * @param dataSourceName
	 * @return Connection
	 * @throws OperationsException 
	 * @throws SQLException 
	 * @throws InterruptedException 
	 * */
	protected synchronized Connection getConnection(String dataSourceName) 
	throws OperationsException, SQLException, InterruptedException {
		if ( null == POOLED_CONNECTION_MAP 
				|| null == POOLED_CONNECTION_MAP.get(dataSourceName) )
			throw new OperationsException("Pool is null.");
		
		if ( null == CONNECTION_POOL_PROPERTY_MAP 
				|| null == DRIVER_PROPERTY_MAP )
			throw new OperationsException("Config is null.");
		
		
		//get resource of DataSource name
		ConnectionPoolProperties cpp = CONNECTION_POOL_PROPERTY_MAP.get(dataSourceName);
		DriverProperties dp = DRIVER_PROPERTY_MAP.get(dataSourceName);
		Vector<PooledConnection> uv = USED_POOLED_CONNECTION_MAP.get(dataSourceName);
		Vector<PooledConnection> fv = POOLED_CONNECTION_MAP.get(dataSourceName);
		
		
		if ( cpp.getMaxPoolSize() <= uv.size() ) {
			
			//If pool is full,the new connection have to
			//wait till daemon thread wake up it.
			CONTAINER_MAP.get(dataSourceName).add(this);
			this.wait(WAIT_MIL_SECONDES);
			
		} 
		return returnConnection(fv, fv, dp);		
	}
	
	/**
	 * Return Connection object
	 * @return Connection
	 * */
	private synchronized Connection returnConnection(
			Vector<PooledConnection> uv,Vector<PooledConnection> fv,DriverProperties dp)
	throws SQLException, OperationsException{
		PooledConnection pc;
		if ( fv.size() > 0 ) {//exist free connection  
			pc = fv.get(0);
			uv.add(pc);
			fv.remove(pc);
			return pc.getConnection();
		}else {//not exist free connection
			pc = DriverConnectionPoolDataSource.
				getInstance(dp).getPooledConnection();
			uv.add(pc);
			return pc.getConnection();
		}
	}
	
	/**
	 * Throw OperationException
	 * @param message String 
	 * @param cause Throwable
	 * */
	private void throwExcpetion(String message,Throwable cause) throws OperationsException{
		throw new OperationsException(message,cause);
	}
	
	/**
	 * GuardThread of connection pool
	 * */
	private class PoolManageThread extends Thread{
		private Vector<SingleRecipientManager> vc;
		private Vector<PooledConnection> uv;
		private ConnectionPoolProperties cpp;
		private static final int WAIT_TIME = 500;
		public PoolManageThread(String dataSourceName) {
			vc = CONTAINER_MAP.get(dataSourceName);
			uv = USED_POOLED_CONNECTION_MAP.get(dataSourceName);
			cpp = CONNECTION_POOL_PROPERTY_MAP.get(dataSourceName);
			this.setDaemon(true);//Marks this thread as a daemon thread 
		}
		public void run(){
			while(true){
				try {
					sleep(WAIT_TIME);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
				while(true) {
					//If computer has not only one CPU,it has to check free connection
					//after finishing current thread.If some connection is free,it must
					//be deal with. 
					if ( !vc.isEmpty() ) {//it has some connections needs deal with,another words,waiting list is not null
						if ( cpp.getMaxPoolSize() > uv.size() ) {//pool is not full
							//deal with the first waiter
							vc.get(0).notify();
							vc.remove(0);//remove it
						} else
							break;
					} else 
						break;
				}
			}
		}
	}
	
	/**
	 * Listener of Pooled connection
	 * */
	private class PooledConnectionListener implements ConnectionEventListener{
		private PooledConnection pc ;
		private Vector<PooledConnection> uv;
		private Vector<PooledConnection> fv;
		public PooledConnectionListener(String dataSourceName) {
			uv = USED_POOLED_CONNECTION_MAP.get(dataSourceName);
			fv = POOLED_CONNECTION_MAP.get(dataSourceName);
		}
		public void connectionClosed(ConnectionEvent event) {
			pc = ((PooledConnection)event.getSource());
			fv.add(pc);
			uv.remove(pc);
		}
		public void connectionErrorOccurred(ConnectionEvent event) {
			event.getSQLException().printStackTrace();	
			pc = ((PooledConnection)event.getSource());
			try {
				pc.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
			uv.remove(pc);	
		}
		
	}
	
}
