/*
    Persistence v1.04 is a easy-use O-R and O-F frame which suit local software.   
    Copyright (C) 2009 www.nonesole.com Jack Lee

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package com.nonesole.persistence.file;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.util.Collection;
import java.util.List;

import com.nonesole.persistence.exception.OperationsException;
import com.nonesole.persistence.file.catalog.CatalogEntry;
import com.nonesole.persistence.file.catalog.CatalogFactory;
import com.nonesole.persistence.file.catalog.ICatalog;
import com.nonesole.persistence.file.search.FileSearchEngine;
import com.nonesole.persistence.file.search.SearchConditions;

public abstract class AbstractFilesOperation implements IFileOperations {

	static CatalogFactory cf = null;
	
	static {
		cf = CatalogFactory.getInstance();
	}
	
	protected AbstractFilesOperation(){}
	
	public void delete(Object obj, URL url) throws OperationsException {
		checkParameters(obj,url,"delete");
		ICatalog cl = this.getCatalog(url);
		delete(obj,cl);
	}

	public Object insert(Object obj, URL url) throws OperationsException {
		checkParameters(obj,url,"insert");
		ICatalog cl = this.getCatalog(url);
		String fullFilePath = getFullFilePath(obj,url);
		insert(obj,cl,fullFilePath);
		return obj;
	}

	public List<Object> query(Object obj, URL url) throws OperationsException {
		//checkParameters(obj,url,"query");
		if ( obj instanceof SearchConditions == false ) {
			throw new OperationsException(
				"Condition's type is failure.");
		}
		ICatalog cl = this.getCatalog(url);
		return FileSearchEngine.getInstance()	
					.lookup((SearchConditions)obj, cl,url);
	}

	public Object update(Object obj, URL url) throws OperationsException {
		checkParameters(obj,url,"update");
		ICatalog cl = this.getCatalog(url);
		String fullFilePath = getFullFilePath(obj, url);
		update(obj,cl,fullFilePath);
		return obj;
	}

	public void batchDelete(List<Object> obj, URL url)
			throws OperationsException {
				checkParameters(obj,url,"batch delete");
				ICatalog cl = this.getCatalog(url);
				for ( Object o : obj ) {
					delete(o,cl);
				}
			}

	public List<Object> batchInsert(List<Object> obj, URL url)
			throws OperationsException {
				checkParameters(obj,url,"batch insert");
				ICatalog cl = this.getCatalog(url);
				String fullFilePath = null;
				for ( Object o : obj ){
					fullFilePath = getFullFilePath(o,url);
					this.insert(o, cl, fullFilePath);
				}
				return obj;
			}

	public List<Object> batchUpdate(List<Object> obj, URL url)
			throws OperationsException {
				checkParameters(obj,url,"batch update");		
				ICatalog cl = this.getCatalog(url);
				String fullFilePath = null;
				for ( Object o : obj ) {
					fullFilePath = getFullFilePath(o,url);
					update(o,cl,fullFilePath);
				}
				return obj;
	}

	/**
	 * Get file nameobj.toString+"ser"
	 * @param obj
	 * @param url 
	 * @return full path
	 * */
	private String getFileName(Object obj) throws OperationsException {
		if ( null == obj )
			throw new OperationsException("File path is null.");
		return new StringBuilder()
			.append(obj.toString())
			.append(this.getExtendFileName())
			.toString();
	}
		
	/**
	 * Delete one object
	 * @param obj 
	 * @param cl 
	 * */
	protected void delete(Object obj,ICatalog cl) 
	throws OperationsException{
		CatalogEntry ce = cl.lookup(obj.toString());
		if ( null == ce )
			return;
		String path = cl.getFolderPath()+"/"+ce.getFileName();
		new File(path).delete();
		cl.unbind(ce);
	}
	/**
	 * Insert one object
	 * @param obj 
	 * @param cl 
	 * @param fullFilepath 
	 * @return Object
	 * */
	protected Object insert(Object obj,ICatalog cl,String fullFilePath) 
	throws OperationsException{
		CatalogEntry ce = cl.lookup(obj.toString());
		if ( null != ce )
			throw new OperationsException(
					"Unable insert the object,it is exist.");
		saveFile(fullFilePath,obj);
		ce = new CatalogEntry();
		ce.setClassName(obj.getClass().getName());
		ce.setFileName(this.getFileName(obj));
		ce.setId(obj.toString());
		cl.bind(ce);
		return obj;
	}
	
	/**
	 * Update one object
	 * @param obj 
	 * @param cl 
	 * @param fullFilepath 
	 * @return Object
	 * */
	protected Object update(Object obj,ICatalog cl,String fullFilePath) 
	throws OperationsException{
		CatalogEntry ce = cl.lookup(obj.toString());
		if ( null == ce )
			throw new OperationsException(
					"Unable update the object,it is not exist.");
		new File(fullFilePath).delete();
		saveFile(fullFilePath,obj);
		return obj;
	}
	
	/**
	 * Save file from object
	 * @param fullFilePath 
	 * @param obj 
	 * */
	protected abstract void saveFile(String fullFilePath,Object obj)
	throws OperationsException;
	
	/**
	 * Get extends name of file
	 * @return name
	 * */
	protected abstract String getExtendFileName();
	
	/**
	 * Get catalog form url
	 * @param url URL
	 * @return ICatalog
	 * @throws OperationsException 
	 * */
	private ICatalog getCatalog(URL url) throws OperationsException {
		try {
			String filePath = url.getFile().replace("%20", " ");
			return cf.reload(filePath).getMemoryCatalog(filePath);
		} catch (IOException e) {
			throw new OperationsException(
					"Getting catalog is failure.",e.fillInStackTrace());
		}
	}
	
	/**
	 * Check parameters
	 * @param obj 
	 * @param url 
	 * @param operation 
	 */
	private void checkParameters(Object obj, URL url, String operation)
			throws OperationsException {
				checkNullParam(obj,url,operation);
				checkObject(obj);
				checkFolder(url);
	}
	

	/**
	 * Check parameter is whether or not null. 
	 * @param Object
	 * @param URL
	 * @param methodName
	 * @throws OperationsException
	 * */
	private void checkNullParam(Object obj, URL url, String methodName)
			throws OperationsException {
				if ( null == obj || null == url )
					throw new OperationsException(
							new StringBuilder()
							.append("parameter of ")
							.append(methodName)
							.append(" operation is null.")
							.toString());
			}

	/**
	 * Check object whether or not valid.
	 * @param obj Object
	 * @throws OperationsException
	 * */
	private void checkObject(Object obj) throws OperationsException {
		if ( obj instanceof Collection ) {
			for ( Object o : Collection.class.cast(obj) ) {
				checkSingleObject(o);
			}
		} else {
			checkSingleObject(obj);
		}
	}

	/**
	 * Check object.
	 * @param obj Object
	 * @throws OperationsException
	 * */
	private void checkSingleObject(Object obj) throws OperationsException {
		Method m = null;
		try {
			//Object which will be checked must have isValid method
			m = obj.getClass().getDeclaredMethod("isValid", new Class[]{});
		} catch (SecurityException e) {
			throw new OperationsException("Get method failure",
					e.fillInStackTrace());
		} catch (NoSuchMethodException e) {
			throw new OperationsException(
				new StringBuilder()
					.append("Class ")	
					.append(obj.getClass().toString())
					.append(" has not got method isValid().")
					.toString()
				);
		} 
		
		try {
			if ( !(Boolean) m.invoke(obj, new Object[]{}) ){
				throw new OperationsException(
						new StringBuilder()
						.append("Object ")
						.append(obj.toString())
						.append(" is not valid. ")
						.toString());
			}
		} catch (IllegalArgumentException e) {
			throw new OperationsException("Reflect isValid method failure",
					e.fillInStackTrace());
		} catch (IllegalAccessException e) {
			throw new OperationsException("Reflect isValid method failure",
					e.fillInStackTrace());
		} catch (InvocationTargetException e) {
			throw new OperationsException("Reflect isValid method failure",
					e.fillInStackTrace());
		}
	}

	/**
	 * URL is whether or not a folder. 
	 * @param url URL
	 * @throws OperationsException 
	 * */
	private void checkFolder(URL url) throws OperationsException {
		if ( !new File(url.getFile().replace("%20", " ")).isDirectory() ) {
			throw new OperationsException("url is not a Folder");
		}
	}

	/**
	 * Get full path
	 * @param obj
	 * @param url 
	 * @return path
	 * */
	private String getFullFilePath(Object obj, URL url)
			throws OperationsException {
				if ( null == url )
					throw new OperationsException("File path is null.");
				return new StringBuilder()
					.append(url.getFile())
					.append("/")
					.append(getFileName(obj)).toString();
	}
	
}