/*
    Persistence v1.04 is a easy-use O-R and O-F frame which suit local software.   
    Copyright (C) 2009 www.nonesole.com Jack Lee

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package com.nonesole.persistence.file.catalog;

import java.io.File;
import java.util.Properties;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.TransformerException;

import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.nonesole.persistence.exception.OperationsException;
import com.nonesole.persistence.tools.XMLHandler;

/**
 * Control Catalog
 * @author JACK LEE
 * @version 1.0 - build in 2009-07-21
 */
public class Catalog implements ICatalog {
		
	//whether or not save changes synchronously
	private boolean isSynchronization = true;
	
	//if not synchronous,how long save changes in circle
	private long circleTime = 60000;
	
	//thread of saving non-synchronous data 
	private Thread thread = null;
	
	private String fileName = null;
	
	private String folderPath = null;
	
	//property for saving data
	private Properties saveProperty = new Properties();
	
	//Class of reading and writing xml file
	private XMLHandler handler = null;
	
	private Document doc = null;
	
	private Node root = null;
	
	/**
	 * class name
	 * */
	protected static final String CLASS_NAME = "className";

	/**
	 * file name
	 * */
	protected static final String FILE_NAME = "fileName";

	/**
	 * id
	 * */
	protected static final String ID = "id";
		
	/**
	 * Construction
	 * @param fileName 
	 * */
	public Catalog(String fileName) throws OperationsException{
		if ( null == fileName || "".equals(fileName) )
			throw new OperationsException("File's name is null");
		this.fileName = fileName.replace("//", "/");
		
		handler = XMLHandler.getInstance();
		saveProperty.put(OutputKeys.ENCODING, "GB2312");
		saveProperty.put(OutputKeys.INDENT, "yes");
		
		File f = new File(fileName);
		if ( !f.isFile() || !f.exists() ) {
			new File(f.getParent()).mkdirs();
			f = null;
			try {
				doc = handler.createDocument();
				doc.appendChild(doc.createElement("Catalog"));
				handler.saveFile(doc, fileName, saveProperty);
			} catch (TransformerException e) {
				throw new OperationsException(
						"Unable create file.",e.fillInStackTrace());
			} catch (DOMException e) {
				throw new OperationsException(
						"Unable create file.",e.fillInStackTrace());
			} catch (ParserConfigurationException e) {
				throw new OperationsException(
						"Unable create file.",e.fillInStackTrace());
			}
		}
		
		try {
			doc = handler.loadFile(fileName);
		} catch (Exception e) {
			throw new OperationsException("Open file error",e.fillInStackTrace());
		}
		root = doc.getDocumentElement();
		
		int lastIndex = this.fileName.lastIndexOf("/");
		this.folderPath = this.fileName.substring(0, lastIndex);
	}
	
	/* (non-Javadoc)
	 * @see com.nonesole.persistence.file.catalog.ICatalog#bind(com.nonesole.persistence.file.catalog.CatalogEntry)
	 */
	public void bind(CatalogEntry ce) 
	throws OperationsException{
		if ( null == ce )
			throw new NullPointerException("CatalogEntry object is null.");
		if ( null != lookup(ce) )
			throw new OperationsException(
					"Unable bind the object,it has been exist");
		
		Element e = doc.createElement("entry");
		root.appendChild(handler.transformToAtrribute(ce, e, doc));		
		if ( this.isSynchronization )
			try {
				save();
			} catch (TransformerException e1) {
				throw new OperationsException(
						e1.getMessage(),e1.fillInStackTrace());
			}
	}
		
	/* (non-Javadoc)
	 * @see com.nonesole.persistence.file.catalog.ICatalog#rebind(com.nonesole.persistence.file.catalog.CatalogEntry)
	 */
	public void rebind(CatalogEntry ce) 
	throws OperationsException{
		if ( null == ce )
			throw new NullPointerException("CatalogEntry object is null.");
				
		Node nd = lookup(ce);
		if ( null == nd )
			throw new NullPointerException("CatalogEntry object is not exist.");
		
		nd.getAttributes()
			.getNamedItem(CLASS_NAME)
			.setNodeValue(ce.getClassName());
		nd.getAttributes()
			.getNamedItem(FILE_NAME)
			.setNodeValue(ce.getFileName());
		if ( this.isSynchronization )
			try {
				save();
			} catch (TransformerException e1) {
				throw new OperationsException(
						e1.getMessage(),e1.fillInStackTrace());
			}
	}
	
	/* (non-Javadoc)
	 * @see com.nonesole.persistence.file.catalog.ICatalog#unbind(com.nonesole.persistence.file.catalog.CatalogEntry)
	 */
	public void unbind(CatalogEntry ce) 
	throws OperationsException{
		if ( null == ce )
			throw new NullPointerException("CatalogEntry object is null.");
		
		Node nd = lookup(ce);
		if ( null == nd )
			throw new NullPointerException("CatalogEntry object is not exist.");
		
		root.removeChild(nd);
		if ( this.isSynchronization )
			try {
				save();
			} catch (TransformerException e1) {
				throw new OperationsException(
						e1.getMessage(),e1.fillInStackTrace());
			}
	}
	
	/**
	 * Search node of CatalogEntry object.
	 * @param ce - CatalogEntry object
	 * @return Node
	 * */
	private Node lookup(CatalogEntry ce){
		if ( null == ce )
			throw new NullPointerException("CatalogEntry object is null.");
		if ( null == ce.getId() || "".equals(ce.getId()) )
			return null;
		NodeList nl = root.getChildNodes();
		Node nd = null;
		for ( int i = 0 ; i < nl.getLength() ; i++ ) {
			nd = nl.item(i);
			if ( "entry".equals(nd.getNodeName()) ) {
				if ( ce.getId().equals(
						nd.getAttributes()
							.getNamedItem(ID)
							.getNodeValue() ) ){
					return nd;
				}
			}
		}
		
		return null;
	}
	
	/* (non-Javadoc)
	 * @see com.nonesole.persistence.file.catalog.ICatalog#lookup(java.lang.String)
	 */
	public CatalogEntry lookup(String id){
		if ( null == id || "".equals(id) )
			return null;
		NodeList nl = root.getChildNodes();
		Node nd = null;
		CatalogEntry ce = null;
		for ( int i = 0 ; i < nl.getLength() ; i++ ) {
			nd = nl.item(i);
			if ( "entry".equals(nd.getNodeName()) ) {
				if ( id.equals(
						nd.getAttributes().getNamedItem(ID)
							.getNodeValue() ) ){
					ce = new CatalogEntry();
					ce.setClassName(nd.getAttributes()
							.getNamedItem(CLASS_NAME).getNodeValue());
					ce.setFileName(nd.getAttributes()
							.getNamedItem(FILE_NAME).getNodeValue());
					ce.setId(nd.getAttributes()
							.getNamedItem(ID).getNodeValue());
					return ce;
				}
			}
		}
		return ce;
	}
	
	/* (non-Javadoc)
	 * @see com.nonesole.persistence.file.catalog.ICatalog#destory()
	 */
	public void destroy() throws InterruptedException{
		this.isSynchronization = true;
		if ( this.thread != null && this.thread.isAlive() )
			wait(circleTime);
		this.thread = null;
		System.gc();
	}
	
	/* (non-Javadoc)
	 * @see com.nonesole.persistence.file.catalog.ICatalog#isSynchronization()
	 */
	public boolean isSynchronization() {
		return isSynchronization;
	}

	/* (non-Javadoc)
	 * @see com.nonesole.persistence.file.catalog.ICatalog#setSynchronization(boolean)
	 */
	public ICatalog setSynchronization(boolean isSynchronization) {
		this.isSynchronization = isSynchronization;
		if ( !isSynchronization ) {
			if ( null == this.thread )
				this.thread = new CircleThread();
			if ( !thread.isAlive() )
				this.thread.start();
		} else {
			if ( !this.thread.isAlive() )
			this.thread = null;
		}
		return this;
	}

	/* (non-Javadoc)
	 * @see com.nonesole.persistence.file.catalog.ICatalog#getCircleTime()
	 */
	public long getCircleTime() {
		return circleTime;
	}

	/* (non-Javadoc)
	 * @see com.nonesole.persistence.file.catalog.ICatalog#setCircleTime(long)
	 */
	public ICatalog setCircleTime(long circleTime) {
		this.circleTime = circleTime;
		return null;
	}

	/**
	 * Save data to xml files
	 * @throws TransformerException 
	 * @throws ParserConfigurationException 
	 * */
	protected void save() throws TransformerException {
		if ( null == doc || null == root ) {
			return;
		}
		handler.saveFile(doc,this.fileName,saveProperty);
	}
	
	/**
	 * Inner thread for save data lazily.
	 * */
	private class CircleThread extends Thread{
		public void run(){
			while(!isSynchronization){
				try {
					sleep(circleTime);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
				try {
					save();
				} catch (TransformerException e) {
					e.printStackTrace();
				}
				
			}
		}
	}

	private int offset = 0;
	
	public boolean hasNext() {
		Node nd = null;
		for ( offset++ ; offset < root.getChildNodes().getLength() ; offset++ ) {
			nd = root.getChildNodes().item(offset);
			if ( "entry".equals(nd.getNodeName()) ) {
				return true;
			}
		}
		return false;
	}

	public CatalogEntry next() {
		Node nd = root.getChildNodes().item(offset);
		CatalogEntry ce = new CatalogEntry();
		ce.setClassName(nd.getAttributes()
				.getNamedItem(CLASS_NAME).getNodeValue());
		ce.setFileName(nd.getAttributes()
				.getNamedItem(FILE_NAME).getNodeValue());
		ce.setId(nd.getAttributes()
				.getNamedItem(ID).getNodeValue());
		return ce;
	}

	public void rewind() {
		offset = 0;		
	}

	public String getFolderPath() {
		return this.folderPath;
	}
		
}
