/*
    Persistence v1.04 is a easy-use O-R and O-F frame which suit local software.   
    Copyright (C) 2009 www.nonesole.com Jack Lee

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package com.nonesole.persistence.file.search;

import java.io.File;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import com.nonesole.persistence.exception.OperationsException;
import com.nonesole.persistence.file.TransformerFactory;
import com.nonesole.persistence.file.catalog.CatalogEntry;
import com.nonesole.persistence.file.catalog.ICatalog;

/**
 * File search engine
 * @author JACK LEE
 * @version 1.0 - build in 2009-07-26
 */
public class FileSearchEngine {

	private static final FileSearchEngine engine = new FileSearchEngine();
	
	private FileSearchEngine(){}
	
	/**
	 * Get engine object
	 * @return FileSearchEngine
	 * */
	public static final FileSearchEngine getInstance(){
		return engine;
	}
	
	/**
	 * Search object by search condition at some URL
	 * @param sc - search condition
	 * @param cl - operation interface of catalog
	 * @param url - file path
	 * */
	public List<Object> lookup(SearchConditions sc,ICatalog cl,URL url) 
	throws OperationsException{
		if ( null == sc || null == cl )
			return null;
		List<Object> l = new ArrayList<Object>(); 
		CatalogEntry ce = null;
		while ( cl.hasNext() ) {
			ce = cl.next();
			if ( this.isNameEquals(sc, ce) 
					& this.isFileEquals(sc, ce)
					& this.isSuitTime(sc, ce, url) ) {
				l.add(getObject(ce,url));
			}
		}
		//Have to rewind after using, 
		//because there is only one ICatalog object in memory.
		cl.rewind();
		return l;
		
	}
	
	/**
	 * Return class names are whether or not equals between 
	 * search condition object and  CatalogEntry object
	 * @param sc search condition
	 * @param ce CatalogEntry
	 * @return true - They equal
	 * */
	private boolean isNameEquals(SearchConditions sc,CatalogEntry ce) {
		if ( null != sc.getClassName() ) {
			if ( !sc.getClassName().equals(ce.getClassName()) ) {
				return false;
			}
		}
		return true;
	}
	
	/**
	 * Return file names are whether or not equals between 
	 * search condition object and  CatalogEntry object
	 * @param sc search condition
	 * @param ce CatalogEntry
	 * @return true - they equal
	 * */
	private boolean isFileEquals(SearchConditions sc,CatalogEntry ce) {
		if ( null != sc.getFileName() && null != ce.getFileName() 
				&& !"".equals(sc.getFileName()) && !"".equals(ce.getFileName()) ) {
			if ( !ce.getFileName().matches(transform(sc.getFileName())) && 
					!ce.getFileName().contains(sc.getFileName()) ) {
				return false;
			}
		}
		return true;
	}
	
	/**
	 * Support *,? and % in string
	 * @param source string
	 * @return has transformed string
	 * */
	private String transform(String source){
		if ( null == source || "".equals(source) )
			return source;
		return source.replace("*", "[\\w|\\W]*")
					 .replace("?", "[\\w|\\W]{1}")
					 .replace("%", "[\\w|\\W]*");
	}
	
	/**
	 * Create time of file is whether or not eligible 
	 * @param sc - SearchConditions
	 * @param ce - CatalogEntry
	 * @return true - time is eligible
	 * */
	private boolean isSuitTime(SearchConditions sc,CatalogEntry ce,URL url) {
		if ( ( (sc.getStartTime() >= 0 
				|| sc.getEndTime() > 0)
				&& sc.getEndTime() >= sc.getStartTime() ) 
				&& null != ce.getFileName() ) {
			String fn = new StringBuilder()
					.append(url.getFile().replace("%20", " "))
					.append("/").append(ce.getFileName())
							.toString().replace("//", "/");
			File f = new File(fn);
			long time = f.lastModified();
			if ( time == 0 )
				return false;
			if ( sc.getStartTime() <= time && sc.getEndTime() >= time) {
				return true;
			} else
				return false;
		}
		return true;
	}
	
	/**
	 * Transform file to object
	 * @param ce - CatalogEntry
	 * @param url - URL
	 * @return object from transforming file 
	 * @throws OperationsException 
	 * */
	private Object getObject(CatalogEntry ce,URL url) throws OperationsException{
		String filePath = new StringBuilder()
							.append(url.getFile().replace("%20", " "))
							.append("//")
							.append(ce.getFileName()).toString();
		return TransformerFactory
					.getInstance()
						.getFileTransformer(filePath)
							.getObject(filePath);
	}
	
}
