/*
    Persistence v1.04 is a easy-use O-R and O-F frame which suit local software.   
    Copyright (C) 2009 www.nonesole.com Jack Lee

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package com.nonesole.persistence.file.txt;

import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Properties;

import com.nonesole.persistence.exception.OperationsException;
import com.nonesole.persistence.file.AuailableCharsets;
import com.nonesole.persistence.file.FileWrapper;
import com.nonesole.persistence.file.IFileTransformer;
import com.nonesole.persistence.tools.FileToolkit;
import com.nonesole.persistence.tools.ReflectToolkit;

/**
 * Text object transformer
 * @author JACK LEE
 * @version 1.0 - build in 2009-07-26
 */
public class TXTTransformer  implements IFileTransformer {

	private final static String FLAG;//Partition flag of Object properties in text file
	
	private final static String PROPERTY_FILE = "txt.properties";//configuration file
	
	private final static TXTTransformer TRANSFORMER;
	
	private final static String CHARSET;
	
	private final static String URI_FILE_HEAD = "file:";
	
	static {
		Properties property = new Properties();
		try {
			property.load(TXTTransformer.class.getResourceAsStream(PROPERTY_FILE));
		} catch (IOException e) {
			new OperationsException("Can not load properties file.");
		} 
		if ( null == property.getProperty("flag") )
			FLAG = "<!-- flag -->";
		else
			FLAG = property.getProperty("flag");
		
		if ( null == property.getProperty("character") )
			CHARSET = AuailableCharsets.UTF_8;
		else
			CHARSET = property.getProperty("character");
		
		TRANSFORMER = new TXTTransformer();
	}
	
	private TXTTransformer(){}
	
	
	public static TXTTransformer getInstance(){
		return TRANSFORMER;
	}
	
	
	public Object getContent(Object obj) throws OperationsException{
		if ( null == obj )
			return null;
		StringBuilder sb = new StringBuilder();
		Class<?> clazz = obj.getClass();
		Field[] fields = clazz.getDeclaredFields();
		sb.append("[class=")
			.append(clazz.getName())
			.append("]")
			.append(FileToolkit.LINE_SEPARATOR);
		Object value = null;
		for ( Field f : fields ) {
			int modifier = f.getModifiers();
			if ( Modifier.isFinal(modifier) ||
					(Modifier.isStatic(modifier) && Modifier.isPublic(modifier) ) ) {
				;//no code here
			} else {		
				value = ReflectToolkit.getFieldValue(f, obj);
				sb.append(getFlag(clazz,f))
					.append(value)
					.append(FileToolkit.LINE_SEPARATOR);}
		}
		return sb.toString();
	}

	/**
	 * Get flags like partition flag and property flag
	 * @param clazz Class of object 
	 * @param f Field Object
	 * @return StringBuilder
	 * */
	private StringBuilder getFlag(Class<?> clazz,Field f){
		return new StringBuilder()
		  .append(FLAG)
		  .append(FileToolkit.LINE_SEPARATOR)
		  .append(getPropertyFlag(clazz,f));
	}
	

	/**
	 * Get property flag
	 * @param clazz Class of object
	 * @param f Field
	 * @return StringBuilder
	 * */
	private StringBuilder getPropertyFlag(Class<?> clazz,Field f){
		return new StringBuilder()
		  .append("[property=")
		  .append(clazz.getSimpleName())
		  .append(".")
		  .append(f.getName())
		  .append("]")
		  .append(FileToolkit.LINE_SEPARATOR);
	}
	
	/**
	 * Get object's class from content of file
	 * @param fileContent
	 * @return class of object
	 * @throws OperationsException 
	 * */
	private Class<?> getObjectClass(String fileContent) throws OperationsException{
		String className =	fileContent.substring(
				fileContent.indexOf("[class=") + "[class=".length(),
					fileContent.indexOf("]"));
		try {
			return Class.forName(className);
		} catch (ClassNotFoundException e) {
			throw new OperationsException(
					"Class is not exist.",e.fillInStackTrace());
		}
	}
	
	/**
	 * Remove the head of file and get the real content
	 * @param fileContent 
	 * @param headRegex - head of file
	 * @return content of file
	 * @throws OperationsException 
	 * */
	private String removeFileHead(String fileContent) throws OperationsException{
		String headRegex = "\\[class=[\\w|.]+\\]";
		String[] realContent = fileContent.split(headRegex); 
		if ( realContent.length > 2 )
			throw new OperationsException("File's head is wrong.");
		return realContent[1];
	}
	
	/**
	 * Get all the paragraphs with partition flag
	 * @param fileContent ļ
	 * @return 
	 * */
	private String[] getParagraphs(String fileContent) throws OperationsException{
		return removeFileHead(fileContent).split(FLAG);
	}
		
	public Object getObject(String fileName) throws OperationsException{
		if ( null == fileName || "".equals(fileName) ){
			throw new OperationsException("File name is null.");
		}
		
		FileWrapper fw = null;
		String fileContent = null;
		try {
			fileName = fileName.replace("%20", " ");
			fw = new FileWrapper(new URI(URI_FILE_HEAD + fileName).toURL());
			fw.setCharset(CHARSET);
			fileContent = fw.readAll();
		} catch (MalformedURLException e) {
			throw new OperationsException(
					"File can not be read.",e.fillInStackTrace());
		} catch (IOException e) {
			throw new OperationsException(
					"File can not be read.",e.fillInStackTrace());
		} catch (URISyntaxException e) {
			throw new OperationsException(
					"File can not be read.",e.fillInStackTrace());
		}
		
		Class<?> clazz = this.getObjectClass(fileContent);
		Object obj = null;
		try {
			obj = clazz.newInstance();
		} catch (InstantiationException e) {
			throw new OperationsException(
					"Can not reflect object.",e.fillInStackTrace());
		} catch (IllegalAccessException e) {
			throw new OperationsException(
					"Can not reflect object.",e.fillInStackTrace());
		}
		Field[] fields = obj.getClass().getDeclaredFields();
		String[] content =  getParagraphs(fileContent);
		
		String propertyFlag = null;
		String param = null;
		for ( String str : content ){
			for ( Field field : fields ){
				propertyFlag = getPropertyFlag(clazz,field).toString();
				if ( str.indexOf(propertyFlag) > -1 ) {
					param = str.substring(propertyFlag.length()).trim();
					ReflectToolkit.setFieldValue(
							field.getName(),
							field.getType(), 
							obj, 
							param);
				}
			}
		}
				
		return obj;
	}

}
