/*
    Persistence v1.04 is a easy-use O-R and O-F frame which suit local software.   
    Copyright (C) 2009 www.nonesole.com Jack Lee

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package com.nonesole.persistence.jndi;

import java.util.Enumeration;
import java.util.Hashtable;

import javax.naming.Binding;
import javax.naming.CompositeName;
import javax.naming.Context;
import javax.naming.InvalidNameException;
import javax.naming.Name;
import javax.naming.NameAlreadyBoundException;
import javax.naming.NameClassPair;
import javax.naming.NameNotFoundException;
import javax.naming.NameParser;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.NotContextException;
import javax.naming.OperationNotSupportedException;

/**
 * Persistence Context Class
 * @author JACK LEE
 * @version 1.0 - build in 2009-07-07
 */
public class PersistenceContext implements Context {

	private Hashtable<String,Object> myEnv;
    private Hashtable<String,Object> bindings = new Hashtable<String,Object>();
    private static NameParser myParser = new PersistenceNameParser();
    
    /**
     * Construction
     * @param environment - Hashtable<String,Object>
     * */
	@SuppressWarnings("unchecked")
	PersistenceContext(Hashtable<?,?> environment) {
        if (environment != null)
			myEnv = (Hashtable<String,Object>)(environment.clone());
		else
			myEnv = null;
    }
	
	public Object addToEnvironment(String propName, Object propVal)
			throws NamingException {
		if (myEnv == null) {
            myEnv = new Hashtable<String,Object>(5, 0.75f);
        } 
        return myEnv.put(propName, propVal);
	}

	public void bind(Name name, Object obj) throws NamingException {
        if ( name == null )
        	throw new NullPointerException("Cannot bind null name");
        else
        	bind(name.toString(), obj);
	}

	public void bind(String name, Object obj) throws NamingException {
		if (name.equals("")) {
            throw new InvalidNameException("Cannot bind empty name");
        }
        bindings.put(name, obj);
	}

	public void close() throws NamingException {
		myEnv = null;
        bindings = null;
	}

	public Name composeName(Name name, Name prefix) throws NamingException {
		Name result = (Name)(prefix.clone());
        result.addAll(name);
        return result;
	}

	public String composeName(String name, String prefix)
			throws NamingException {
		Name result = composeName(new CompositeName(name),
				new CompositeName(prefix));
		return result.toString();
	}

	public Context createSubcontext(Name name) throws NamingException {
        return createSubcontext(name.toString());
	}

	public Context createSubcontext(String name) throws NamingException {
		throw new OperationNotSupportedException(
    		"PersistenceCtx does not support subcontexts");
	}

	public void destroySubcontext(Name name) throws NamingException {
        destroySubcontext(name.toString());
	}

	public void destroySubcontext(String name) throws NamingException {
		throw new OperationNotSupportedException(
        	"FlatCtx does not support subcontexts");
	}

	public Hashtable<?, ?> getEnvironment() throws NamingException {
		if (myEnv == null) {
            return new Hashtable<String,Object>(3, 0.75f);
        } else {
            return (Hashtable<?,?>)myEnv.clone();
        }
	}

	public String getNameInNamespace() throws NamingException {
		return ""; 
	}

	public NameParser getNameParser(Name name) throws NamingException {
		return myParser;
	}

	public NameParser getNameParser(String name) throws NamingException {
        return getNameParser(name.toString());
	}

	public NamingEnumeration<NameClassPair> list(Name name)
			throws NamingException {
        return list(name.toString());
	}

	public NamingEnumeration<NameClassPair> list(String name)
			throws NamingException {
		if (name.equals("")) {
            return new PersistenceNames(bindings.keys());
        } 

        // Perhaps 'name' names a context
        Object target = lookup(name);
        if (target instanceof Context) {
            return ((Context)target).list("");
        }
        throw new NotContextException(name + " cannot be listed");
	}

	public NamingEnumeration<Binding> listBindings(Name name)
			throws NamingException {
        return listBindings(name.toString());
	}

	public NamingEnumeration<Binding> listBindings(String name)
			throws NamingException {
		if (name.equals("")) {
            // listing this context
            return new PersistenceBindings(bindings.keys());
        } 

        // Perhaps 'name' names a context
        Object target = lookup(name);
        if (target instanceof Context) {
            return ((Context)target).listBindings("");
        }
        throw new NotContextException(name + " cannot be listed");
	}

	public Object lookup(Name name) throws NamingException {
		return lookup(name.toString());
	}

	public Object lookup(String name) throws NamingException {
		if (name.equals("")) {
            // Asking to look up this context itself.  Create and return
            // a new instance with its own independent environment.
            return (new PersistenceContext(myEnv));
        }
        Object answer = bindings.get(name);
        if (answer == null) {
            throw new NameNotFoundException(name + " not found");
        }
        return answer;
	}

	public Object lookupLink(Name name) throws NamingException {
		return lookupLink(name.toString());
	}

	public Object lookupLink(String name) throws NamingException {
        return lookup(name);
	}

	public void rebind(Name name, Object obj) throws NamingException {
		bind(name.toString(), obj);
	}

	public void rebind(String name, Object obj) throws NamingException {
		if (name.equals("")) {
            throw new InvalidNameException("Cannot bind empty name");
        }
        bindings.put(name, obj);		
	}

	public Object removeFromEnvironment(String propName) throws NamingException {
		if (myEnv == null)
            return null;

        return myEnv.remove(propName);
	}

	public void rename(Name oldName, Name newName) throws NamingException {
		rename(oldName.toString(), newName.toString());
	}

	public void rename(String oldName, String newName) throws NamingException {
		if (oldName.equals("") || newName.equals("")) {
            throw new InvalidNameException("Cannot rename empty name");
        }

        // Check if new name exists
        if (bindings.get(newName) != null) {
            throw new NameAlreadyBoundException(newName +
                                                " is already bound");
        }

        // Check if old name is bound
        Object oldBinding = bindings.remove(oldName);
        if (oldBinding == null) {
            throw new NameNotFoundException(oldName + " not bound");
        }

        bindings.put(newName, oldBinding);
	}

	public void unbind(Name name) throws NamingException {
		unbind(name.toString());
	}

	public void unbind(String name) throws NamingException {
		if (name.equals("")) {
            throw new InvalidNameException("Cannot unbind empty name");
        }
        bindings.remove(name);		
	}
	
	// Class for enumerating name/class pairs
    class PersistenceNames implements NamingEnumeration<NameClassPair> {
        Enumeration<String> names;

        PersistenceNames (Enumeration<String> enumeration) {
            this.names = enumeration;
        }

        public boolean hasMoreElements() {
            return names.hasMoreElements();
        }

        public boolean hasMore() throws NamingException {
            return hasMoreElements();
        }

        public NameClassPair nextElement() {
            String name = (String)names.nextElement();
            String className = bindings.get(name).getClass().getName();
            return new NameClassPair(name, className);
        }

        public NameClassPair next() throws NamingException {
            return nextElement();
        }
        public void close() {
        }
    }

    // Class for enumerating bindings
    class PersistenceBindings implements NamingEnumeration<Binding> {
        Enumeration<String> names;

        PersistenceBindings (Enumeration<String> enumeration) {
            this.names = enumeration;
        }

        public boolean hasMoreElements() {
            return names.hasMoreElements();
        }

        public boolean hasMore() throws NamingException {
            return hasMoreElements();
        }

        public Binding nextElement() {
            String name = (String)names.nextElement();
            return new Binding(name, bindings.get(name));
        }

        public Binding next() throws NamingException {
            return nextElement();
        }
        public void close() {
        }
    }
	
}
