/*
    Persistence v1.04 is a easy-use O-R and O-F frame which suit local software.   
    Copyright (C) 2009 www.nonesole.com Jack Lee

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package com.nonesole.persistence.sql;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.Blob;
import java.sql.Connection;
import java.sql.Date;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.Map;
import java.util.Vector;

import com.nonesole.persistence.exception.OperationsException;
import com.nonesole.persistence.sql.parser.SqlConfigMapping;
import com.nonesole.persistence.sql.parser.SqlConfigParser;

/**
 * Deal with object by prepared sql.
 * @author JACK LEE
 * @version 1.0 - build in 2009-07-29
 */
public class PreparedStatementDML {
	
	private static final PreparedStatementDML PSDML;
	static {
		PSDML = new PreparedStatementDML();
	}
	private PreparedStatementDML(){}
	
	/**
	 * Get PreparedStatementDML object
	 * @return PreparedStatementDML
	 * */
	public static PreparedStatementDML getInstance(){
		return PSDML;
	}
		
	/**
	 * Get O-R mapping
	 * @param obj - Object
	 * @return SqlConfigMapping
	 * */
	private SqlConfigMapping getSqlConfigMapping(Object obj) 
	throws OperationsException{
		if ( null == obj )
			throw new OperationsException("Object is null.");
		
		Map<String,SqlConfigMapping> config 
			= SqlConfigParser.getIntance().getConfigMap();
		if ( null == config 
				|| null == config.get(obj.getClass().getName()) )
			throw new OperationsException("Sql config is failure.");
		
		return (SqlConfigMapping) config.get(obj.getClass().getName());
	}
	
	/**
	 * Get fields of Object
	 * @param obj Object
	 * @return Field[]
	 * */
	private Field[] getFields(Object obj) throws OperationsException{
		Class<?> c = obj.getClass();
		java.lang.reflect.Field[] fields = c.getDeclaredFields();
		if ( fields == null )
			throw new OperationsException("Fields is null.");
		return fields;
	}
				
	/**
	 * Get prepared statement by conn and sql
	 * @param conn Connection
	 * @return Statement
	 * */
	private PreparedStatement getStatement(Connection conn,String sql) 
	throws OperationsException{
		if ( null == conn ) {
			throw new OperationsException(
					"Connection is null.");
		}
		try {
			return conn.prepareStatement(sql);			
		} catch (SQLException e) {
			throw new OperationsException(
					"Creating statement is failure.",e.fillInStackTrace());
		}
	}
		
	/**
	 * Set values to prepared statement to finish integral sql.
	 * Just blob not clob.
	 * @param st PreparedStatement which has prepared sql
	 * @param values to '?' in sql
	 * @param type of values
	 * @param blob's position in valus 
	 * @param startPosition start position of values 
	 * */
	private void setValuesToPreparedStatement(
			PreparedStatement st,
			Vector<Object> values,
			Vector<Class<?>> fTypes,
			Vector<Integer> blobPosition,
			int startPosition)
	throws OperationsException {
		try {
			Class<?> type = null;
			int i = 0;
			int position = (startPosition<=0)?1:startPosition;//Get start position
			for ( Object obj : values ) {
				type = fTypes.get(i);//Get type
				if ( boolean.class == type || Boolean.class == type ) {
					obj = (obj==null)?false:obj;
					st.setBoolean(position, (Boolean)obj);
				} else if ( int.class == type || Integer.class == type ) {
					obj = (obj==null)?0:obj;
					st.setInt(position, (Integer)obj);
				} else if ( long.class == type || Long.class == type ) {
					obj = (obj==null)?0:obj;
					st.setLong(position, (Long)obj);
				} else if ( float.class == type || Float.class == type ) {
					obj = (obj==null)?0.0f:obj;
					st.setFloat(position, (Float)obj);
				} else if ( double.class == type || Double.class == type ) {
					obj = (obj==null)?0.0d:obj;
					st.setDouble(position, (Double)obj);
				} else if ( short.class == type || Short.class == type ) {
					obj = (obj==null)?0:obj;
					st.setShort(position, (Short)obj);
				} else if ( Byte.class == type || byte.class == type ) {
					obj = (obj==null)?new Byte(""):obj;
					st.setByte(position, (Byte)obj);
				} else if ( String.class == type ) {
					obj = (obj==null)?"":obj;
					st.setString(position, (String)obj);
				} else if ( type == Date.class  ){
					obj = (obj==null)?new Date(System.currentTimeMillis()):obj;
					st.setDate(position, (Date)obj);
				} else if ( type == Time.class  ){
					obj = (obj==null)?new Time(System.currentTimeMillis()):obj;
					st.setTime(position, (Time)obj);
				} else if ( type == Timestamp.class  ){
					obj = (obj==null)?new Timestamp(System.currentTimeMillis()):obj;
					st.setTimestamp(position, (Timestamp)obj);
				} else if ( type == Character.class || type == char.class ){
					obj = (obj==null)?"":obj;
					st.setString(position, String.valueOf(obj));
				} else if ( type == BigDecimal.class  ){
					obj = (obj==null)?0:obj;
					st.setBigDecimal(position, (BigDecimal)obj);
				} else if ( type == BigInteger.class ){
					obj = (obj==null)?0:obj;
					st.setLong(position, ((BigInteger)obj).longValue());
				} else if ( type == byte[].class ){
					setBolbValue(st,blobPosition,position);//deal with blob
				} else {
					obj = (obj==null)?type.newInstance():obj;
					st.setObject(position, obj);
				} 
				i++;
				position++;
			}
		} catch (SQLException e) {
			throw new OperationsException(
					"Setting value for preparement is failure.",
					e.fillInStackTrace());
		} catch (InstantiationException e) {
			throw new OperationsException(
					"Setting value for preparement is failure.",
					e.fillInStackTrace());
		} catch (IllegalAccessException e) {
			throw new OperationsException(
					"Setting value for preparement is failure.",
					e.fillInStackTrace());
		}
	}
	
	/**
	 * Set values to '?' of where
	 * @param st PreparedStatement object
	 * @param values to '?'
	 * @param type of values
	 * @param startPosition start position of values 
	 * */
	private void setValueToWhere(
			PreparedStatement st,
			Vector<Object> values,
			Vector<Class<?>> fTypes,
			int startPosition)
	throws OperationsException {
		this.setValuesToPreparedStatement(st, values, fTypes, null,startPosition);
	}
	
	/**
	 * Deal with blob value
	 * @param st PreparedStatement object
	 * @param blob's position in values
	 * @param fieldPosition field position in fields
	 * */
	private void setBolbValue(
			PreparedStatement st,Vector<Integer> blobPosition,int fieldPosition) 
	throws OperationsException {
		if ( null != blobPosition ) {
			//Blob <-> byte[],This frame do not deal with blob in other types
			try {
				st.setBinaryStream(fieldPosition, 
						new DataInputStream(new ByteArrayInputStream("".getBytes())) ,
						0);
			} catch (SQLException e) {
				throw new OperationsException(
						"Found mistakes when set Blob field.",
						e.fillInStackTrace());
			}
			blobPosition.add(fieldPosition);
		} 
	}
	
	/**
	 * Get query sql whose id as condition.
	 * @param conn Connection Object
	 * @param obj object for query
	 * @return SQL
	 * */
	private String getSQL(Connection conn,Object obj) 
	throws OperationsException{
		SqlConfigMapping scp = getSqlConfigMapping(obj);
		Field[] fields = getFields(obj);
		String name;//field name
		Class<?> type = null;//field type
		StringBuilder set = new StringBuilder();//sql like 'field=value'
		StringBuilder where = new StringBuilder(); //where sql
		StringBuilder temp = null;
		Object value;//value of string type
		
		Vector<String> key = scp.getPrimeryKey();
		for( Field field : fields ){
			name = field.getName();
			type = field.getType();	
			int modifier = field.getModifiers();
			if ( Modifier.isFinal(modifier) ||
					(Modifier.isStatic(modifier) && Modifier.isPublic(modifier) ) ) {
				//public + static,no need set and get method
				continue;
			} 				
			if ( key.contains(name) ) {
				temp = new StringBuilder();
				//get value
				value = SQLDML.getInstance().invoke(field, obj);
				temp.append(name).append("=");
				//whether or not use '' around character 
				SQLDML.getInstance().handleCharacter(type,temp,value);
				//join set sub sentence and where sub setnece 
				set.append(temp).append(",");
				where.append(temp).append(" and ");
			}
		} 
		//delete last ','
		int point = set.lastIndexOf(",");
		set.delete(point, point+1);
		point = where.lastIndexOf(" and ");
		where.delete(point, point+6);
		
		StringBuilder sql = new StringBuilder();
		sql.append("select * from")
			.append(" ")
			.append(scp.getTableName())
			.append(" ")
			.append("where")
			.append(" ")
			.append(where);
		
		return sql.toString();
	}
	
	/**
	 * Get connection which contains insert sql  
	 * @param conn Connection Object
	 * @param obj object needs insert
	 * @return connection which contains insert sql
	 * */
	public Connection getInsertConnection(Connection conn,Object obj) 
	throws OperationsException{
		return getInsertConnection(conn,obj,getFields(obj));
	}
	
	/**
	 * Get connection which contains update sql  
	 * @param conn Connection Object
	 * @param obj object needs update
	 * @return connection which contains update sql
	 * */
	public Connection getUpdateConnection(Connection conn,Object obj) 
	throws OperationsException{
		return getUpdateConnection(conn,obj,getFields(obj));
	}
	
	/**
	 * Get connection which contains insert sql  
	 * @param conn Connection Object
	 * @param obj object needs insert
	 * @param fields fields needs insert
	 * @return connection which contains insert sql
	 * */
	public Connection getInsertConnection(Connection conn,Object obj,Field[] fields) 
	throws OperationsException{
			
		SqlConfigMapping scp = getSqlConfigMapping(obj);
		
		if ( null == fields )
			fields = getFields(obj);
		
		String name;
		Class<?> type = null;
		StringBuilder dbField = new StringBuilder();
		StringBuilder values = new StringBuilder(); 
		Vector<Object> v = new Vector<Object>();
		Vector<Class<?>> fTypes = new Vector<Class<?>>();
		
		for( Field field : fields ){
			name = field.getName();
			type = field.getType();
			int modifier = field.getModifiers();
			if ( Modifier.isFinal(modifier) ||
					(Modifier.isStatic(modifier) && Modifier.isPublic(modifier) ) ) {
				continue;
			} 
			dbField.append(name).append(",");
			v.add(SQLDML.getInstance().invoke(field, obj));
			fTypes.add(type);
			values.append("?,");			
		} 
		int point = dbField.lastIndexOf(",");
		dbField.delete(point, point+1);
		point = values.lastIndexOf(",");
		values.delete(point, point+1);
				
		StringBuilder sql = new StringBuilder();
		
		sql.append("insert into")
			.append(" ")
			.append(scp.getTableName())
			.append("(")
			.append(dbField.toString())
			.append(") values (")
			.append(values.toString())
			.append(")");
		
		//Get Statement
		PreparedStatement st = getStatement(conn,sql.toString());
		Vector<Integer> blobPosition = new Vector<Integer>();
		setValuesToPreparedStatement(st,v,fTypes,blobPosition,0);
		//if blob is exist,deal with it
		execute(st);	
		saveBlob(conn,obj,blobPosition,v);
		return conn;
	}
	
	/**
	 * Get connection which contains update sql  
	 * @param conn Connection Object
	 * @param obj object needs update
	 * @param fields fields needs update
	 * @return connection which contains update sql
	 * */
	public Connection getUpdateConnection(Connection conn,Object obj,Field[] fields) 
	throws OperationsException{
		
		SqlConfigMapping scp = getSqlConfigMapping(obj);
		if ( null == fields )
			fields = getFields(obj);
		String name;
		Class<?> type = null;
		StringBuilder set = new StringBuilder();
		StringBuilder where = new StringBuilder(); 
		StringBuilder temp = null;
		Object value = null;
		Vector<Object> values = new Vector<Object>();
		Vector<Class<?>> fTypes = new Vector<Class<?>>();
		Vector<String> key = scp.getPrimeryKey();
		Vector<Object> primeryKeyValues = new Vector<Object>();
		Vector<Class<?>> primeryKeyTypes = new Vector<Class<?>>();
		
		for( Field field : fields ){
			name = field.getName();
			type = field.getType();	
			int modifier = field.getModifiers();
			if ( Modifier.isFinal(modifier) ||
					(Modifier.isStatic(modifier) && Modifier.isPublic(modifier) ) ) {
				continue;
			} 	
			temp = new StringBuilder();
			value = SQLDML.getInstance().invoke(field, obj);
			values.add(value);
			fTypes.add(type);
			temp.append(name).append("=").append("?");
			set.append(temp).append(",");
			if ( key.contains(name) ) {
				where.append(temp).append(" and ");
				//Save prepared parameter
				primeryKeyValues.add(value);
				primeryKeyTypes.add(type);
			}
						
		} 
		int point = set.lastIndexOf(",");
		set.delete(point, point+1);
		point = where.lastIndexOf(" and ");
		where.delete(point, point+6);
		
		StringBuilder sql = new StringBuilder();
		sql.append("update")
			.append(" ")
			.append(scp.getTableName())
			.append(" ")
			.append("set")
			.append(" ")
			.append(set)
			.append(" ")
			.append("where")
			.append(" ")
			.append(where);
		
		PreparedStatement st = getStatement(conn,sql.toString());
		Vector<Integer> blobPosition = new Vector<Integer>(); 
		setValuesToPreparedStatement(st,values,fTypes,blobPosition,0);
		setValueToWhere(st, primeryKeyValues, primeryKeyTypes, values.size()+1);
		execute(st);	
		saveBlob(conn,obj,blobPosition,values);
		return conn;
	}
	
	/**
	 * Execute sql
	 * @param st PreparedStatement Object
	 * */
	private void execute(PreparedStatement st) throws OperationsException {
		try {
			//System.out.println(st);
			st.execute();
			st.close();
			st = null;
		} catch (SQLException e) {
			e.printStackTrace();
			throw new OperationsException(
					"Executing PreparedStatement is failure.",e.fillInStackTrace());
		}
	}
	
	/**
	 * Deal with blob field
	 * @param conn Connection Object
	 * @param obj Object needs save
	 * @param blobPosition blob's position in fields
	 * @param v list of values 
	 * */
	private void saveBlob(Connection conn,
			Object obj,
			Vector<Integer> blobPosition,
			Vector<Object> v) throws OperationsException{
		if ( blobPosition.size() < 1 || v.size() < 1 ) {
			return;
		}
		
		try {
			Statement stt = conn.createStatement();
			ResultSet rs = stt.executeQuery(getSQL(conn,obj));
			if (rs.next()) {
				Blob blob;
				int size;
				for ( Integer i : blobPosition ) {
					size = ((byte[])v.get(i-1)).length;
					blob = rs.getBlob(i);
					blob.setBytes(1, (byte[])v.get(i-1), 0, size);
				}				
			}
			rs.close();
			rs = null;
			stt.close();
			stt = null;
		} catch (SQLException e) {
			throw new OperationsException(
					"Executing PreparedStatement is failure.",e.fillInStackTrace());
		}
		
	}
	
}
