/*
    Persistence v1.04 is a easy-use O-R and O-F frame which suit local software.   
    Copyright (C) 2009 www.nonesole.com Jack Lee

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package com.nonesole.persistence.sql;

import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.Connection;
import java.sql.Date;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import com.nonesole.persistence.exception.OperationsException;
import com.nonesole.persistence.sql.parser.SqlConfigMapping;
import com.nonesole.persistence.sql.parser.SqlConfigParser;

/**
 * Provide methods who can transform object to sql.
 * <pre>
 * Example
 * String sql = DML.getInstance().transferToDeleteSql(new JavaBean());
 * DML.getInstance().execute(sql,DataSource.getConnection());
 * </pre>
 * It spent 1.5s on transforming 50,000 rows data which have 5 field.
 * @author JACK LEE
 * @version 1.0 - build in 2009-07-14
 */
public class SQLDML {

	private final static SQLDML dml;
	static {
		dml = new SQLDML();
	}
	private SQLDML(){}
	
	/**
	 * Get SQLDML Object
	 * @return DML Object
	 * */
	public static SQLDML getInstance(){
		return dml;
	}
	
	/**
	 * Get value frome get method
	 * @param field Field Object 
	 * @param obj Object needs do with
	 * @throws OperationsException
	 * */
	protected Object invoke(Field field,Object obj) throws OperationsException{
		int modifier = field.getModifiers();
		if ( Modifier.isFinal(modifier) ||
				(Modifier.isStatic(modifier) && Modifier.isPublic(modifier) ) ) {
			return null;
		} 
		Class<?> fieldType = field.getType();
		String fieldName = field.getName();
		Method m = null;
		Class<?> classType = obj.getClass();
		StringBuilder sb = new StringBuilder();
		if ( fieldType == boolean.class ) {
			//if some method return boolean then needs special deal with 
		    if ( "is".equals(fieldName.substring(0,2)) )//isXXX method
		    	sb.append("is").
				append(fieldName.substring(2, fieldName.length()));
		    else//field name do not contain "is"
				sb.append("is").
				append(fieldName.substring(0,1).toUpperCase()).
				append(fieldName.substring(1, fieldName.length()));
		
		} else//non-boolean-return method
			sb.append("get").
				append(fieldName.substring(0,1).toUpperCase()).
				append(fieldName.substring(1, fieldName.length()));
		try {
			m = classType.getMethod(sb.toString(),new Class[]{});
		} catch (SecurityException e) {
			throw new OperationsException(
					"Getting method failure.",e.fillInStackTrace());
		} catch (NoSuchMethodException e) {
			e.printStackTrace();
			throw new OperationsException(
					"Getting method failure.",e.fillInStackTrace());
		}
		try {
			return m.invoke(obj, new Object[]{});
		} catch (IllegalArgumentException e) {
			throw new OperationsException(
					"Invoking method failure.",e.fillInStackTrace());
		} catch (IllegalAccessException e) {
			throw new OperationsException(
					"Invoking method failure.",e.fillInStackTrace());
		} catch (InvocationTargetException e) {
			throw new OperationsException(
					"Invoking method failure.",e.fillInStackTrace());
		}
	}
	
	/**
	 * Check field whether or not needs '' around it's value
	 * @param fieldType  field type
	 * @param sb string which needs join
	 * @param value value which will join the string
	 * @throws OperationsException
	 * */
	protected void handleCharacter(
			Class<?> fieldType,StringBuilder sb,Object value) 
	throws OperationsException{
		
		if ( fieldType == String.class 
				|| fieldType == Date.class 
				|| fieldType == Time.class
				|| fieldType == Timestamp.class 
				|| fieldType == Character.class
				|| fieldType == char.class)	{
			sb.append("'").append(value).append("'");
		} else if ( boolean.class == fieldType || Boolean.class == fieldType
				 	|| int.class == fieldType || Integer.class == fieldType 
				 	|| long.class == fieldType || Long.class == fieldType
				 	|| float.class == fieldType || Float.class == fieldType
				 	|| double.class == fieldType || Double.class == fieldType
				 	|| short.class == fieldType || Short.class == fieldType
				 	|| Byte.class == fieldType || byte.class == fieldType
				 	|| BigDecimal.class == fieldType 
				 	|| BigInteger.class == fieldType ) {
			sb.append(value);
		} else if ( fieldType.isArray() ) {
			Class<?> arrayType = fieldType.getComponentType();
			int length = Array.getLength(value);
			if ( arrayType.isPrimitive() ) {
				//if value is primitive
				for ( int i = 0 ; i < length ; i++ ) {
					sb.append(Array.get(value, i));
				}
			} else {
				for ( int i = 0 ; i < length ; i++ ) {
					sb.append(Array.get(value, i).toString());
				}
			}
			
		}else {
			sb.append(value.toString());
		}
	}
	
	/**
	 * Transform object to delete sql
	 * @param obj Object needs transform
	 * @return String
	 * */
	public String transferToDeleteSql(Object obj) throws OperationsException{
		if ( null == obj )
			throw new OperationsException("Object is null.");
		
		Map<String,SqlConfigMapping> config 
			= SqlConfigParser.getIntance().getConfigMap();
		if ( null == config 
				|| null == config.get(obj.getClass().getName()) )
			throw new OperationsException("Sql config is failure.");
		
		SqlConfigMapping scp = 
			(SqlConfigMapping) config.get(obj.getClass().getName());
		
		Class<?> c = obj.getClass();
		java.lang.reflect.Field[] fields = c.getDeclaredFields();
		if ( fields == null )
			throw new OperationsException("Fields is null.");
					
		String name;
		Object value;
		Class<?> type = null;
		StringBuilder where = new StringBuilder(); 
		StringBuilder temp = null;
		
		Vector<String> key = scp.getPrimeryKey();
		for( Field field : fields ){
			name = field.getName();
			type = field.getType();	
			int modifier = field.getModifiers();
			if ( Modifier.isFinal(modifier) ||
					(Modifier.isStatic(modifier) && Modifier.isPublic(modifier) ) ) {
				//no set and get method if field is public + static 
				continue;
			} 			
			temp = new StringBuilder();
			value = this.invoke(field, obj);
			temp.append(name).append("=");
			handleCharacter(type,temp,value);
			
			if ( key.contains(name) ) {
				where.append(temp).append(" and ");
			}
						
		} 
		int point = where.lastIndexOf(" and ");
		where.delete(point, point+6);
		
		StringBuilder sql = new StringBuilder();
		
		sql.append("delete from")
			.append(" ")
			.append(scp.getTableName())
			.append(" ")
			.append("where")
			.append(" ")
			.append(where);
		
		return sql.toString();
	}
	
	/**
	 * /**
	 * Transform object to update sql
	 * @param obj Object needs transform
	 * @return String
	 * */
	public String transferToUpdateSql(Object obj) throws OperationsException{
		if ( null == obj )
			throw new OperationsException("Object is null.");
		
		Map<String,SqlConfigMapping> config 
			= SqlConfigParser.getIntance().getConfigMap();
		if ( null == config 
				|| null == config.get(obj.getClass().getName()) )
			throw new OperationsException("Sql config is failure.");
		
		SqlConfigMapping scp = 
			(SqlConfigMapping) config.get(obj.getClass().getName());
		
		Class<?> c = obj.getClass();
		java.lang.reflect.Field[] fields = c.getDeclaredFields();
		if ( fields == null )
			throw new OperationsException("Fields is null.");
					
		String name;
		Object value;
		Class<?> type = null;
		StringBuilder set = new StringBuilder();
		StringBuilder where = new StringBuilder(); 
		StringBuilder temp = null;
		
		Vector<String> key = scp.getPrimeryKey();
		for( Field field : fields ){
			name = field.getName();
			type = field.getType();	
			int modifier = field.getModifiers();
			if ( Modifier.isFinal(modifier) ||
					(Modifier.isStatic(modifier) && Modifier.isPublic(modifier) ) ) {
				continue;
			} 
			temp = new StringBuilder();
			value = this.invoke(field, obj);
			temp.append(name).append("=");
			handleCharacter(type,temp,value);
			
			set.append(temp).append(",");
			if ( key.contains(name) ) {
				where.append(temp).append(" and ");
			}
						
		} 
		int point = set.lastIndexOf(",");
		set.delete(point, point+1);
		point = where.lastIndexOf(" and ");
		where.delete(point, point+6);
		
		StringBuilder sql = new StringBuilder();
		
		sql.append("update")
			.append(" ")
			.append(scp.getTableName())
			.append(" ")
			.append("set")
			.append(" ")
			.append(set)
			.append(" ")
			.append("where")
			.append(" ")
			.append(where);
		
		return sql.toString();
	}
	 
	/**
	 * Transform object to insert sql
	 * @param obj Object needs transform
	 * @return String
	 * */
	public String transferToInsertSql(Object obj) throws OperationsException{
		if ( null == obj )
			throw new OperationsException("Object is null.");
		
		Class<?> c = obj.getClass();
		java.lang.reflect.Field[] fields = c.getDeclaredFields();
		if ( fields == null )
			throw new OperationsException("Fields is null.");
		
		String name;
		Object value;
		Class<?> type = null;
		StringBuilder dbField = new StringBuilder();
		StringBuilder values = new StringBuilder();
		for( Field field : fields ){
			name = field.getName();
			type = field.getType();
			int modifier = field.getModifiers();
			if ( Modifier.isFinal(modifier) ||
					(Modifier.isStatic(modifier) && Modifier.isPublic(modifier) ) ) {
				continue;
			} 
			dbField.append(name).append(",");
			value = this.invoke(field, obj);
			handleCharacter(type,values,value);
			values.append(",");			
		} 
		int point = dbField.lastIndexOf(",");
		dbField.delete(point, point+1);
		point = values.lastIndexOf(",");
		values.delete(point, point+1);
		
		
		//Get current configurations
		Map<String,SqlConfigMapping> config 
			= SqlConfigParser.getIntance().getConfigMap();
		if ( null == config 
				|| null == config.get(obj.getClass().getName()) )
			throw new OperationsException("Sql config is failure.");
		
		//Get O-R mapping
		SqlConfigMapping scp = 
			(SqlConfigMapping) config.get(obj.getClass().getName());
		
		//sql
		StringBuilder sql = new StringBuilder();
		sql.append("insert into")
			.append(" ")
			.append(scp.getTableName())
			.append("(")
			.append(dbField.toString())
			.append(") values (")
			.append(values.toString())
			.append(")");
		
		return sql.toString();
	}
	
	/**
	 * excute sql
	 * @param sql 
	 * @param conn Connection
	 * @throws OperationsException
	 * */
	protected void execute(String sql,Connection conn) throws OperationsException {
		if ( null == sql || "".equals(sql) )
			throw new OperationsException("Sql is null");
		
		sql.replaceAll("'", "''");
		Statement stmt = null;
		try {
			stmt = (Statement) conn.createStatement();
			stmt.execute(sql);
		} catch (SQLException e) {
			try {
				if ( null != conn && !conn.isClosed() )
					try {
						conn.rollback();
					} catch (SQLException e1) {
						throw new OperationsException(
								"Executing sql is failure.",e.fillInStackTrace());
					}
			} catch (SQLException e1) {
				throw new OperationsException(
						"Executing sql is failure.",e1.fillInStackTrace());
			}
			throw new OperationsException(
					"Executing sql is failure.",e.fillInStackTrace());
		} finally {
			if ( conn != null ) {
				try {
					if ( stmt != null ) {
						stmt.close();
						stmt = null;
					}
					conn.close();
					conn = null;
				} catch (SQLException e) {
					throw new OperationsException(
							"Executing sql is failure.",e.fillInStackTrace());
				}
			}
		}
		
	}
	
	/**
	 * execute sqls 
	 * @param sql 
	 * @param conn Connection
	 * @throws OperationsException
	 * */
	protected void beathExecute(List<String> list,Connection conn) throws OperationsException {
		if ( null == list )
			throw new OperationsException("Sql is null");
		
		try {
			conn.setAutoCommit(false);
		} catch (SQLException e) {
			throw new OperationsException(
					"Setting auto commit is failure.",e.fillInStackTrace());
		}
				
		Statement stmt = null;
		try {
			stmt = (Statement) conn.createStatement();
			for ( String sql : list ) {
				if ( null == sql || "".equals(sql) )
					throw new OperationsException("Sql is null");
				sql.replaceAll("'", "''");
				stmt.execute(sql);
			}
			conn.commit();			
		} catch (SQLException e) {
			if ( null != conn )
				try {
					conn.rollback();
				} catch (SQLException e1) {
					throw new OperationsException(
							"Rollback transaction is failure.",e.fillInStackTrace());
				}
			throw new OperationsException(
					"Committing transaction is failure.",e.fillInStackTrace());
		} finally {
			if ( conn != null ) {
				try {
					if ( stmt != null ) {
						stmt.close();
						stmt = null;
					}
					conn.setAutoCommit(true);
					conn.close();
					conn = null;
				} catch (SQLException e) {
					throw new OperationsException(
							"Executing sql is failure.",e.fillInStackTrace());
				}
			}
		}
		
	}
		
}
