/*
    Persistence v1.04 is a easy-use O-R and O-F frame which suit local software.   
    Copyright (C) 2009 www.nonesole.com Jack Lee

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package com.nonesole.persistence.sql.parser;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import com.nonesole.persistence.exception.OperationsException;
import com.nonesole.persistence.main.FrameConfigConstruction;
import com.nonesole.persistence.main.FrameEnvService;
import com.nonesole.persistence.sql.SQLConstant;
import com.nonesole.persistence.tools.ReflectToolkit;
import com.nonesole.persistence.tools.XMLHandler;

/**
 * Parser of sql configuration
 * <pre>
 * Example
 *   Default path:
 *   SqlConfigParser scp = 
 *   	SqlConfigParser.getInstance();
 *   Vector<SqlConfigParser> v = scp.getConfig();
 *   
 *   Outside define the path:
 *   SqlConfigParser scp = 
 *   	SqlConfigParser.getInstance().setConfigFilePath("c:\\a.xml");
 *   Vector<SqlConfigParser> v = scp.getConfig();
 * </pre>
 * @author JACK LEE
 * @version 1.0 - build in 2009-07-14
 */
public class SqlConfigParser {

	//Default configuration file path
	private static final String FILE_PATH;
	private Vector<SqlConfigMapping> configs;
	private String filePath = null;
	private Map<String,SqlConfigMapping> configsMap;
	
	private static SqlConfigParser scp;
	private SqlConfigParser(){
		configs = new Vector<SqlConfigMapping>();
		configsMap = new HashMap<String,SqlConfigMapping>();
	}
	static {
		FILE_PATH = FrameEnvService
						.getInstance()
						.getProperty(FrameConfigConstruction.DEFAULT_ORM_CONFIG);
		scp = new SqlConfigParser();
	}
	
	/**
	 * Get SqlConfigParser object
	 * */
	public static SqlConfigParser getIntance(){
		return scp;
	}
	
	/**
	 * Set configuration file's path(Optional)
	 * <pre>
	 * Example
	 *   SqlConfigParser scp = 
	 *   	SqlConfigParser.getInstance().setConfigFilePath("c:\\a.xml");
	 * </pre>
	 * @param filePath 
	 * @return SqlConfigParser 
	 * */
	public SqlConfigParser setConfigFilePath(String filePath){
		this.filePath = filePath;
		return this;
	}
	
	/**
	 * Return configuration
	 * @return array of SqlConfigMapping 
	 * @throws OperationsException 
	 * */
	public Vector<SqlConfigMapping> getConfig() throws OperationsException{
		if ( configs.size() == 0 )
			parser();
		return configs;
	}
	
	/**
	 * Return O-R mappings
	 * @return SqlConfigMapping ( class name is key )
	 * @throws OperationsException 
	 * */
	public Map<String,SqlConfigMapping> getConfigMap() throws OperationsException{
		if ( configsMap.size() == 0 )
			parser();
		return configsMap;
	}
	
	/**
	 * Configuration Parser 
	 * @throws Exception 
	 * */
	private void parser() throws OperationsException{
		Document doc;
		try {
			if ( null == this.filePath )
				this.filePath = FILE_PATH;
			doc = XMLHandler.getInstance().loadFile(this.filePath);
		} catch (Exception e) {
			throw new OperationsException("Parsing is failure.",
											e.fillInStackTrace());
		}
		Node root = doc.getDocumentElement();
		List<Node> l = XMLHandler.getInstance().getNodes(root, 
				SQLConstant.OBJECT);//Get list of nodes
		
		if ( null == l )
			throw new OperationsException("Config file has no config information.");
		
		SqlConfigMapping scf = null;
		for ( Node n : l ) {
			try {
				scf = parseParameter(n);
			} catch (Exception e) {
				throw new OperationsException("Parsing is failure.",
						e.fillInStackTrace());
			}
			configs.add(scf);
			configsMap.put(scf.getClassName(), scf);
		}
	}
	
	/**
	 * Pares xml file and get SqlConfingMapping object
	 * @param n Node object
	 * @return SqlConfingMapping object
	 * */
	private static final SqlConfigMapping parseParameter(Node n) throws Exception{
		Map<String,Object> m = new HashMap<String,Object>();
		
		NamedNodeMap nnm = n.getAttributes();
		if ( null == nnm
				|| null == nnm.getNamedItem(SQLConstant.OBJECT_BEAN)
				|| null == nnm.getNamedItem(SQLConstant.OBJECT_TABLE))
			return null;
		
		//transform node to map,property of node as key of map and value of node as value of map
		Field[] fields =  SqlConfigMapping.class.getDeclaredFields();
		String[] primeryKeys;
		Vector<String> v = new Vector<String>();
		for ( Field field : fields ) {
			if ( Vector.class == field.getType() ) {
				primeryKeys = nnm.getNamedItem(field.getName())
					.getNodeValue()
					.split(SQLConstant.PRIMERY_SEPERATOR);
				for ( String primery : primeryKeys )
					v.add(primery);
				m.put(field.getName(),v);
			} else
				m.put(field.getName(),
					nnm.getNamedItem(field.getName()).getNodeValue());
		}
		
		return (SqlConfigMapping)ReflectToolkit
					.parseMapToObject(m,SqlConfigMapping.class);
	}
}
