/*
    Persistence v1.04 is a easy-use O-R and O-F frame which suit local software.   
    Copyright (C) 2009 www.nonesole.com Jack Lee

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package com.nonesole.persistence.tools;

import java.beans.XMLDecoder;
import java.beans.XMLEncoder;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.RandomAccessFile;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Properties;

/**
 * File Tools
 * @author JACK LEE
 * @version 1.0 - build in 2008-03-29
 * */
public class FileToolkit {
     
	/**
	 * Line separator
	 * */
	public final static String LINE_SEPARATOR = "\r\n";
	
   /**
    * Load file from full path
    * @param fullPath 
    * @return File Object
    * @throws IOException 
    * */
    public static final File loadFile(String fullPath) throws IOException {
    	if ( FormatToolkit.isNull(fullPath) ) {
    		throw new IOException("The path is null.");
    	}
    	File file = new File(fullPath);
    	if ( !file.isFile() ) {
    		throw new IOException("Can't open the file by using the path.");
    	}
    	return file;
    }
        
    /**
     * Read file to FileInputStream
     * @param fullPath 
     * @return FileInputStream object
     * @throws IOException 
     * */   
    public static final FileInputStream loadFileToInputStream(String fullPath) 
    throws IOException {
    	return new FileInputStream(loadFile(fullPath));
    }
        
    /**
     * Read file to serialization Object
     * @param fullPath 
     * @return Object serialization object
     * @throws IOException 
     * @throws ClassNotFoundException
     * */   
    public static final Object loadFileToObject(String fullPath) 
    throws IOException, ClassNotFoundException {
    	if ( null == fullPath ) return null;
    	fullPath = fullPath.replace("%20", " ");
    	File f = new File(fullPath);
    	if ( !f.exists() || !f.isFile() ) return null;
    	FileInputStream fis = new FileInputStream(f);
    	ObjectInputStream ois = new ObjectInputStream(fis);
    	Object obj = ois.readObject();
    	ois.close();
    	fis.close();
    	ois = null;
    	fis = null;
   	 	f = null;
    	return obj;
    }
    
    /**
     * Read file to Properties Object
     * @param fullPath 
     * @return Properties
     * @throws IOException 
     * @throws ClassNotFoundException
     * */   
    public static final Properties loadProperties(String fullPath) 
    throws IOException, ClassNotFoundException {
    	Properties prop = new Properties();
    	prop.load(loadFileToInputStream(fullPath));
    	return prop;
    }
        
    /**
     * Read file to String
     * @param fullPath 
     * @return string
     * @throws IOException 
     * @throws ClassNotFoundException
     * */   
    public static final String loadFileToString(String fullPath) 
    throws IOException, ClassNotFoundException {
    	FileReader read = new FileReader(fullPath);
    	BufferedReader  br = new BufferedReader(read);
    	String str = "";
    	while(br.ready())
    		str += br.readLine() + FileToolkit.LINE_SEPARATOR;
    	return str;
    }
    
    /**
     * Read file to byte[]
     * @param fullPath
     * @return byte[]
     * @throws IOException 
     * @throws ClassNotFoundException
     * */   
    public static final byte[] loadFileToByteArray(String fullPath) 
    throws IOException, ClassNotFoundException {
    	File file = FileToolkit.loadFile(fullPath);
    	FileInputStream fis = new FileInputStream(file);
    	long size = file.length();
    	byte[] bt = null;
    	if ( size > Integer.MAX_VALUE ){
    		throw new IOException(
    				"File size is greater than array upper limit(=Integer.MAX_VALUE).");
    	} else {
    		bt = new byte[(int)size ];
    		fis.read(bt);
    	}
    	return bt;
    }
    
    /**
     * Create new file
     * @param fullPath
     * @return File
     * @throws IOException
     * */
    public static final File createNewFile(String fullPath) 
    throws IOException{
    	if ( null == fullPath ) return null;
    	fullPath = fullPath.replace("%20", " ");
    	File file = new File(fullPath);
	   	if ( file.isDirectory() ) {
	   		throw new IOException("The path is folder's,not a file's.");
	   	} else if ( !file.isFile() ) {
	   		File folder = new File(file.getParent());
	   		folder.mkdirs();
	   		folder = null;
	   		file.createNewFile();
	   	}
	   	return file;
    }
    
    /**
     * Save file form object
     * @param fullPath
     * @param object needs save
     * @throws IOException 
     * */
     public static final void saveFileFromObject(String fullPath,Object obj) 
     throws IOException {
    	 isNull(fullPath,obj);
    	 File f = createNewFile(fullPath);
    	 FileOutputStream fos = new FileOutputStream(f);
    	 ObjectOutputStream oos = new ObjectOutputStream(fos);
    	 oos.writeObject(obj);
    	 oos.flush();
    	 oos.close();
    	 fos.close();
    	 oos = null;
    	 fos = null;
    	 f = null;
    	 
     }
        
    /**
     * Save file form string
     * @param fullPath
     * @param string
     * @throws IOException 
     * */
	public static final void saveFileFromString(String fullPath,String str) 
	throws IOException {
		isNull(fullPath,str);
		saveFileFromByteArray(fullPath,str.getBytes());
	}
      
    /**
     * Save file form byte[]
     * @param fullPath
     * @param byte[]
     * @throws IOException 
     * */
	public static final void saveFileFromByteArray(String fullPath,byte[] array) 
	throws IOException {
		isNull(fullPath,array);
		File f = createNewFile(fullPath);
     	FileOutputStream fOut = new FileOutputStream(f);
     	ByteArrayOutputStream byteOut = new ByteArrayOutputStream(1024);
     	byteOut.write(array);
     	byteOut.writeTo(fOut);
 		fOut.close();
 		byteOut.close();
 		fOut = null;
 		byteOut = null;
 		f = null;
	}
	
	/**
	 * Load file array in dictionary
	 * @param fullPath
	 * @param suffix of file name like ".txt"
	 * @return Files array
	 * @throws IOException
	 * */
	public static final File[] loadFiles(String fullPath,String suffix) 
	throws IOException {
		isNull(fullPath,suffix);
		fullPath = fullPath.replace("%20", " ");
		File file = new File(fullPath);
		if ( !file.isDirectory() ) {
			file = new File(file.getParent());
		}
		return file.listFiles(new FileFilter(suffix));
	}
	
	/**
	 * Get files' name array in dictionary
	 * @param fullPath
	 * @param suffix of file name like ".txt"
	 * @return File names array
	 * @throws IOException
	 * */
	public static final String[] getFileNames(String fullPath,String suffix) 
	throws IOException {
		isNull(fullPath,suffix);
		fullPath = fullPath.replace("%20", " ");
		File file = new File(fullPath);
		if ( !file.isDirectory() ) {
			file = new File(file.getParent());
		}
		return file.list(new FileFilter(suffix));
	}
	
	/**
	 * Check file path and object whether are null. 
	 * @param fullPath
	 * @param object
	 * @throws IOException
	 * */
	private static void isNull(String fullPath,Object obj) 
	throws IOException {
		if ( FormatToolkit.isNull(fullPath) ) {
			throw new IOException("The path is null.");
      	}
     	if ( FormatToolkit.isNull(obj) ) {
       		throw new IOException("The parameter is null.");
       	}
	}
	
	/**  
	 * Read file to PrintWriter Object for output
     * @param filepath
     * @return  PrintWriter object
     */  
    public static final PrintWriter getPrintWriter(String filepath) {  
    	try {
			return new PrintWriter(
						new OutputStreamWriter(   
							new FileOutputStream(filepath, true)));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		return null;
    }
    
    /**  
     * insert content into file  
     * @param file  
     * @param content  
     */  
    public static final void insertContentByPrintWriter(String file, String conent) {   
    	PrintWriter pw = getPrintWriter(file);
    	pw.write(conent);
    	pw.close();
    	pw = null;
    } 
	
	/**  
     * insert content into file  
     * @param fileName  
     * @param content  
     */  
    public static final void insertContentByBufferWriter(String file, String conent) {   
        BufferedWriter out = null;   
        try {   
            out = new BufferedWriter(new OutputStreamWriter(   
                    new FileOutputStream(file, true)));   
            out.write(conent);   
        } catch (Exception e) {   
            e.printStackTrace();   
        } finally {   
            try {   
                out.close();   
            } catch (IOException e) {   
                e.printStackTrace();   
            }   
        }   
    }   
  
    /**  
     * insert content into file  
     * @param fileName  
     * @param content  
     */  
    public static final void insertContentByFileWriter(String fileName, String content) {   
        try {   
            FileWriter writer = new FileWriter(fileName, true);   
            writer.write(content);   
            writer.close();   
        } catch (IOException e) {   
            e.printStackTrace();   
        }   
    }   
  
    /**  
     * insert content into file  
     * @param fileName  
     * @param content   
     */  
    public static final void insertContentByRondomAccessFile(String fileName, String content) {   
        try {     
            RandomAccessFile randomFile = new RandomAccessFile(fileName, "rw");   
            long fileLength = randomFile.length();
            randomFile.seek(fileLength);   
            randomFile.writeBytes(content);   
            randomFile.close();   
        } catch (IOException e) {   
            e.printStackTrace();   
        }   
    }   

	/**
	 * Get root path of project
	 * @return URL root path
	 * */
	public static final URL getDefaultRootPath(){
		String s = FileToolkit.class.getResource("/").toString();
		String[] path = s.split("/");
		StringBuilder sb = new StringBuilder();
		for ( int i = 0 ; i < path.length - 1 ;i++ ){
			sb.append(path[i]).append("/");
		}
				
		try {
			URI uri = new URI(sb.toString());
			return uri.toURL();
		} catch (URISyntaxException e) {
			e.printStackTrace();
		} catch (MalformedURLException e) {
			e.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Save JavaBean to File
	 * @param obj JavaBean
	 * @param url URL of file
	 * */
	public static final void saveFileToXML(Object obj,URL url) 
	throws FileNotFoundException{
		saveFileToXML(obj,url.getFile().replace("%20", " "));
	}
	
	/**
	 * Read JavaBean from File
	 * @param url URL of file
	 * @return JavaBean
	 * */
	public static final Object loadFileToXML(URL url)
	throws FileNotFoundException{
		return loadFileToXML(url.getFile());
	}
	
	/**
	 * Save JavaBean to File
	 * @param obj JavaBean
	 * @param url path of file (String)
	 * */
	public static final void saveFileToXML(Object obj,String url) 
	throws FileNotFoundException{
		XMLEncoder e = new XMLEncoder(
                new BufferedOutputStream(
                    new FileOutputStream(url)));
		e.writeObject(obj);
		e.close();
		e = null;
	}
	
	/**
	 * Read JavaBean from File
	 * @param url path of file (String)
	 * @return JavaBean
	 * */
	public static final Object loadFileToXML(String url)
	throws FileNotFoundException{
		XMLDecoder d = new XMLDecoder(
                new BufferedInputStream(
                    new FileInputStream(url)));
		Object result = d.readObject();
		d.close();
		d = null;
		return result;
	}
	
}
