/*
    Persistence v1.04 is a easy-use O-R and O-F frame which suit local software.   
    Copyright (C) 2009 www.nonesole.com Jack Lee

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package com.nonesole.persistence.tools;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.nonesole.persistence.exception.OperationsException;
import com.nonesole.persistence.file.AuailableCharsets;

/**
 * XML Tools
 * @author JACK LEE
 * @version 1.0 - build in 2009-07-21
 */
public class XMLHandler {

	private static XMLHandler handler;
	
	static {
		handler = new XMLHandler();
	}
	
	private XMLHandler(){}
	
	/**
	 * Get XMLHandler Object
	 * @return XMLHandler Object
	 * */
	public static XMLHandler getInstance(){
		return handler;
	}
		
	/**
	 * Read XML File and return Document object
	 * @param filePath String
	 * @return document Document
	 * @throws ParserConfigurationException 
	 * @throws IOException 
	 * @throws SAXException 
	 * */
	public final Document loadFile(String filePath) 
	throws ParserConfigurationException, SAXException, IOException{
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		DocumentBuilder db = dbf.newDocumentBuilder();
		Document doc = null;
		doc = db.parse(filePath);
		return doc;
	}
	

	/**
	 * Save Document Object to XML File<br>
	 * <pre>
	 * Example
	 *   Properties property = new Properties();
	 *   property.put(OutputKeys.ENCODING, "GB2312");
	 *   property.put(OutputKeys.INDENT, "yes");
	 *   XMLHandler.getInstance()
	 *   	.saveFile(new Document(),"D:\\test.xml",property);
	 * </pre>
	 * @param doc Document
	 * @param filePath 
	 * @param property Properties configuration set
	 * @throws TransformerException 
	 * @throws Exception
	 * */
	public final void saveFile(Document doc,String filePath,Properties property)
	throws TransformerException {
	    TransformerFactory tFactory = TransformerFactory.newInstance();
	    Transformer transformer = tFactory.newTransformer();
	    DOMSource source = new DOMSource(doc);
	    StreamResult result = new StreamResult(new File(filePath.replace("%20", " ")));
	    Object[] keys = property.keySet().toArray();
	    for ( int i = 0 ; i < keys.length ; i++ ) {
	    	transformer.setOutputProperty(
	    			(String)keys[i],(String)property.get(keys[i]));
	    }	    
	    transformer.transform(source, result);
	}	
	
	/**
	 * Transform StringBuilder to Document 
	 * @param xmlString - StringBuilder
	 * @return Document
	 * @throws ParserConfigurationException 
	 * @throws IOException 
	 * @throws SAXException 
	 */
	public final Document parse(StringBuilder xmlString) 
		throws ParserConfigurationException, SAXException, IOException {
		DocumentBuilder builder = createDocumentBuilder();
		Reader reader = new StringReader(xmlString.toString());
		InputSource is = new InputSource(reader);
		return builder.parse(is);
	}

	/**
	 * byte[] to Document
	 * @param xml - the byte[] array object;
	 * @return Document Object;
	 * @throws ParserConfigurationException 
	 * @throws IOException 
	 * @throws SAXException 
	 */
	public final Document parse(byte[] xml) 
		throws ParserConfigurationException, SAXException, IOException {
		DocumentBuilder builder = createDocumentBuilder();
		ByteArrayInputStream input = new ByteArrayInputStream(xml);
		return builder.parse(input);
	}
	
	/**
	 * Create new Document
	 * @return DOM Document Object;
	 * @throws ParserConfigurationException 
	 */
	public final Document createDocument() 
		throws ParserConfigurationException {
		DocumentBuilder builder = createDocumentBuilder();
		return builder.newDocument();
	}

	/**
	 * Transform Object to Element<br>
	 * Field of Object is attribute of element. 
	 * If it exists some objects implements 
	 * java.util.Collection and java.util.Map,
	 * the sub element will be treated as child 
	 * node of current element.
	 * <pre>
	 * Example
	 * 	Document doc = XMLHandler.createDocument();
	 * 	CatalogEntry ce = new CatalogEntry();
	 * 	Element e = doc.createElement("entry");
	 * 	doc.appendChild(XMLHandler.getInstance().transformToAtrribute(ce, e, doc));
	 * XML File
	 * 	<person name="tom" age=10 sex="boy">
	 * 		<houses>
	 * 			<house address="21 Ado road BeiJing"/>
	 * 			<house address="122 Wei Min Street BeiJing"/>
	 * 		</houses>
	 * 	</person>
	 * </pre>
	 * 
	 * @param obj Object needs transform
	 * @param ele Empty element
	 * @param doc Document object for creating new element
	 * @return result element
	 * */
	public final Element transformToAtrribute(Object obj,Element ele,Document doc) 
	throws DOMException, OperationsException{
		if ( null == obj || null == ele ) {
			throw new NullPointerException((null==obj)?"Object":"Element" + " is null");
		}
		Field[] fields = obj.getClass().getDeclaredFields();
		Object value = null;
		Element child = null;
		Object[] keys = null;
		for ( Field f : fields ) {
			int modifier = f.getModifiers();
			if ( Modifier.isFinal(modifier) ||
					(Modifier.isStatic(modifier) && Modifier.isPublic(modifier) ) ) {
				continue;
			} 
			value = ReflectToolkit.getFieldValue(f, obj);
			if ( null != value  ) {
				if ( value instanceof java.util.Collection ) {
					for ( Object o : (java.util.Collection<?>)value ) {
						child = doc.createElement(o.toString());
						ele.appendChild(this.transformToAtrribute(o, child, doc));
					}
				} else if ( value instanceof java.util.Map<?, ?> ){
					keys = ((java.util.Map<?, ?>)value).keySet().toArray();
					for ( Object o : keys ) {
						child = doc.createElement(o.toString());
						ele.appendChild(
								this.transformToAtrribute(
										((java.util.Map<?, ?>)value).get(o), child, doc));
					}
				} else
					ele.setAttribute(f.getName(),value.toString());
			}
			else
				ele.setAttribute(f.getName(), "");
		}
		return ele;			
		
	}
	
	/**
	 * Transform Document to String
	 * @param doc Document
	 * @param charset CharSet
	 * @return result
	 */
	public final String parseToString(Document doc,String charset) {
		if ( null == doc )
			throw new NullPointerException("Document object is null.");
		if ( null == charset || "".endsWith(charset) )
			charset = AuailableCharsets.UTF_8;
		
		TransformerFactory  tf  =  TransformerFactory.newInstance();
		Transformer t;
		String xmlStr = "";
		try {
			t = tf.newTransformer();
			t.setOutputProperty("encoding",charset);
			ByteArrayOutputStream  bos  =  new  ByteArrayOutputStream();
			t.transform(new DOMSource(doc), new StreamResult(bos));
			xmlStr = bos.toString();
		} catch (TransformerConfigurationException e) {
			return "";
		} catch (TransformerException e) {
			return "";
		}
		return xmlStr;
	}
	
	/**
	 * Craete DocumentBuilder Object
	 * @throws ParserConfigurationException 
	 * */
	private final DocumentBuilder createDocumentBuilder() 
		throws ParserConfigurationException{
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		return builder;
	}
	
	/**
	 * Transform Document to String
	 * @param doc Document
	 * @return result
	 */
	public final String xmlDocumentToString(Document doc,String charactor) {
		TransformerFactory  tf  =  TransformerFactory.newInstance();
		Transformer t;
		String xmlStr = "";
		try {
			t = tf.newTransformer();
			t.setOutputProperty("encoding",charactor);
			ByteArrayOutputStream  bos  =  new  ByteArrayOutputStream();
			t.transform(new DOMSource(doc), new StreamResult(bos));
			xmlStr = bos.toString();
		} catch (TransformerConfigurationException e) {
			return "";
		} catch (TransformerException e) {
			return "";
		}
		return xmlStr;
	}
	
	/**
	 * Transform Document to StringBuilder
	 * @param doc Document
	 * @return StringBuilder
	 */
	public final StringBuilder xmlDocumentToStringBuilder(Document doc,String character) {
		return new StringBuilder(xmlDocumentToString(doc,character));
	}
	
	/**
	 * Transform Document to byte[]
	 * @param doc Document
	 * @return byte[] 
	 * */
	public final byte[] parseToStream(Document doc,String character){
		return xmlDocumentToString(doc,character).getBytes();
	}
	
	/**
	 * Get list of node whose name is given
	 * @param node father node
	 * @param nodeName child node name
	 * @return list of child nodes
	 * */
	public final List<Node> getNodes(Node node,String nodeName) {
		if ( null == node )
			return null;
		NodeList nl = node.getChildNodes();
		List<Node> r = new ArrayList<Node>();
		for ( int i = 0 ; i < nl.getLength() ; i++  ) {
			if ( nodeName.equals(nl.item(i).getNodeName()) ) {
					r.add(nl.item(i));
			}
		}
		return r;
	}		
}
